"""Read the experimental CPMG data."""

import importlib
import math

import numpy as np


def read_profiles(path, filenames, details, model, included=None, excluded=None):
    """Read the CPMG profiles."""
    experiment_type = details["type"].split(".")
    details["name"] = name_experiment(details)
    experiment_module = importlib.import_module(
        ".".join(
            ["chemex.experiments", experiment_type[0], "profiles", experiment_type[1]]
        )
    )

    error = details.get("error", "file")
    if error not in {"file", "auto"}:
        print("Warning: The 'error' option should either be 'file' or ")
        print("'auto'. Using the default 'file' option.")
        error = "file"

    error_values = []

    if included is None:
        included = filenames.keys()
    else:
        included = [_.lower() for _ in included]

    if excluded is None:
        excluded = []
    else:
        excluded = [_.lower() for _ in excluded]

    Profile = getattr(experiment_module, "Profile")

    dtype = [("ncycs", "<f8"), ("intensities", "<f8"), ("intensities_err", "<f8")]

    profiles = []

    for profile_name, filename in filenames.items():

        full_path = path / filename

        measurements = np.loadtxt(full_path, dtype=dtype)

        profile = Profile(profile_name, measurements, details, model)

        is_included = profile_name in included
        is_not_excluded = profile_name not in excluded

        if is_included and is_not_excluded:
            profiles.append(profile)

        if error == "auto":
            error_values.append(estimate_noise(profile))

    if error == "auto":
        error_value = np.mean(error_values)
        for profile in profiles:
            profile.err[:] = error_value

    ndata = sum(len(profile.val) for profile in profiles)

    return profiles, ndata


def name_experiment(experiment_details=None):
    """Generate a unique name for the experiment."""
    if not experiment_details:
        experiment_details = dict()

    if "name" in experiment_details:
        name = experiment_details["name"].strip().replace(" ", "_")

    else:
        exp_type = experiment_details["type"].replace(".", "_")
        h_larmor_frq = float(experiment_details["h_larmor_frq"])
        temperature = float(experiment_details["temperature"])
        time_t2 = float(experiment_details["time_t2"])

        name = "{:s}_{:.0f}ms_{:.0f}MHz_{:.0f}C".format(
            exp_type, time_t2 * 1e3, h_larmor_frq, temperature
        ).lower()

    return name


def correction(n):
    """Calculate correction factor for noise estimate."""
    k = n // 2

    if n == 2 * k:
        factor = (
            math.sqrt(math.pi * (2 * k - 1) / 2.0)
            * math.factorial(2 * k - 2)
            / ((2 ** (2 * k - 2)) * (math.factorial(k - 1) ** 2))
        )
    else:
        factor = (
            math.sqrt(k / math.pi)
            * (2 ** (2 * k - 1))
            * (math.factorial(k - 1) ** 2)
            / math.factorial(2 * k - 1)
        )

    return factor


def estimate_noise(profile):
    """Estimate the uncertainty in CPMG data points (i.e., R2eff)."""
    intensity_dict = {}

    for ncyc, intensity in zip(profile.ncycs, profile.val):
        intensity_dict.setdefault(ncyc, []).append(intensity)

    std_list = []
    for duplicates in intensity_dict.values():
        n_duplicates = len(duplicates)
        if n_duplicates > 1:
            std_list.append(np.std(duplicates, ddof=1) * correction(n_duplicates))

    if std_list:
        error = np.mean(std_list)
    else:
        error = np.mean(profile.err)

    return error



"""13CO - Pure Anti-phase Carbonyl 13C CPMG

Analyzes carbonyl chemical exchange that is maintained as anti-phase
magnetization throughout the CPMG block. This results in lower intrinsic
relaxation rates and therefore better sensitivity. The calculations use a 12x12,
2-spin exchange matrix:

[ COx(a), COy(a), COz(a), 2COxNz(a), 2COyNz(a), 2COzNz(a),
  COx(b), COy(b), COz(b), 2COxNz(b), 2COyNz(b), 2COzNz(b)]

Notes
-----

Because of the length of the shaped pulses used during the CPMG blocks, off-
resonance effects are taken into account only for the 90-degree pulses that
create COxNz before the CPMG and COzNz after the CPMG. The calculation is
designed explicitly for analyzing the Kay laboratory pulse sequence:

CO_CPMG_SCFilter_x00_dfh1

And can be run with or without sidechain CO inversion via the Inv_CO flag for
uniformly 13C-labeled proteins.

Reference
---------

Journal of Biomolecular NMR (2008) 42, 35-47

"""

import numpy as np

from chemex.experiments.cpmg import cpmg_profile
from chemex.spindynamics import basis, default

EXP_DETAILS = {
    "carrier": {"type": float},
    "time_t2": {"type": float},
    "pw90": {"type": float},
    "time_equil": {"type": float, "default": 0.0},
    "sidechain_flg": {"type": str, "default": "n"},
    "taucc": {"type": float, "default": 9.09e-3},
}


class Profile(cpmg_profile.CPMGProfile):
    """TODO: class docstring."""

    def __init__(self, name, measurements, exp_details, model):
        super().__init__(name, measurements, exp_details, model)

        self.exp_details = self.check_exp_details(exp_details, expected=EXP_DETAILS)

        self.liouv = basis.Liouvillian(
            system="ixyzsz",
            state_nb=self.model.state_nb,
            atoms=self.peak.atoms,
            h_larmor_frq=self.conditions["h_larmor_frq"],
            equilibrium=False,
        )

        self.liouv.carrier_i = self.exp_details["carrier"]
        self.liouv.w1_i = 2.0 * np.pi / (4.0 * self.exp_details["pw90"])

        self.detect = self.liouv.detect["2izsz_a"]

        self.t_cps = {
            ncyc: self.exp_details["time_t2"] / (4.0 * ncyc)
            for ncyc in self.ncycs[self.ncycs > 0]
        }

        self.t_neg = -2.0 * self.exp_details["pw90"] / np.pi

        self.delays = [
            self.t_neg,
            self.exp_details["time_equil"],
            self.exp_details["taucc"],
        ]
        self.delays.extend(self.t_cps.values())

        self.sidechain_flg = self.exp_details["sidechain_flg"].lower().startswith("y")

        self.map_names, self.params = default.create_params(
            basis=self.liouv,
            model=self.model,
            nuclei=self.peak.names,
            conditions=self.conditions,
            hn_ap_constraints=True,
        )

        for name, full_name in self.map_names.items():
            if name.startswith(("dw", "r2_i_a")):
                self.params[full_name].set(vary=True)

    def calculate_unscaled_profile(self, **parvals):
        """Calculate the intensity in presence of exchange after a CEST block.

        Parameters
        ----------
        pb : float
            Fractional population of state B.
        kex_ab : float
            Exchange rates between states A and B in /s.
        dw_i_ab : float
            Chemical shift difference between states A and B in rad/s.
        r1_i_a : float
            Longitudinal relaxation rate of states A in /s.
        r2_i_a : float
            Transverse relaxation rate of state A in /s.
        dr2_i_ab : float
            Transverse relaxation rate difference between states A and B in /s.
        cs_i_a : float
            Resonance position of state A in ppm.

        Returns
        -------
        out : float
            Intensity after the CEST block

        """

        self.liouv.update(**parvals)

        # Calculation of all the needed propagators
        pulses = self.liouv.pulses_90_180_i()
        delays = dict(zip(self.delays, self.liouv.delays(self.delays)))

        p_180pmy = 0.5 * (pulses["180py"] + pulses["180my"])  # +/- phase cycling

        # Simulate the CPMG block as function of ncyc
        mag0 = self.liouv.compute_mag_eq(term="2izsz", **parvals)

        profile = []

        if self.sidechain_flg:
            p_flip = p_180pmy
        else:
            p_flip = (
                pulses["90my"]
                @ delays[self.exp_details["taucc"]]
                @ p_180pmy
                @ delays[self.exp_details["taucc"]]
                @ pulses["90py"]
            )

        for ncyc in self.ncycs:

            if ncyc == 0:

                mag = (
                    self.detect
                    @ delays[self.exp_details["time_equil"]]
                    @ pulses["90py"]
                    @ p_flip
                    @ pulses["90py"]
                    @ mag0
                )

            else:

                p_cpx = np.linalg.matrix_power(
                    delays[self.t_cps[ncyc]]
                    @ self.liouv.perfect180["ix"]
                    @ delays[self.t_cps[ncyc]],
                    int(ncyc),
                )

                mag = (
                    self.detect
                    @ delays[self.exp_details["time_equil"]]
                    @ pulses["90py"]
                    @ delays[self.t_neg]
                    @ p_cpx
                    @ delays[self.t_neg]
                    @ p_flip
                    @ delays[self.t_neg]
                    @ p_cpx
                    @ delays[self.t_neg]
                    @ pulses["90py"]
                    @ mag0
                )

            profile.append(np.float64(self.liouv.collapse(mag)))

        return np.asarray(profile)

    def ncycs_to_nu_cpmgs(self, ncycs=None):
        """Calculate the pulsing frequency, v(CPMG), from ncyc values."""

        if ncycs is None:
            ncycs = self.ncycs

        return ncycs / self.exp_details["time_t2"]



"""1H - Pure Anti-phase Proton CPMG

Analyzes amide proton chemical exchange that is maintained as anti-phase
magnetization throughout the CPMG block. This results in lower intrinsic
relaxation rates and therefore better sensitivity. The calculations use a 12x12,
2-spin exchange matrix:

[ Hx(a), Hy(a), Hz(a), 2HxNz(a), 2HyNz(a), 2HzNz(a),
  Hx(b), Hy(b), Hz(b), 2HxNz(b), 2HyNz(b), 2HzNz(b)]

Note
----
Off resonance effects are taken into account. The calculation is designed
explicitly for analyzing the Lewis Kay pulse sequence:

H1_CPMG_Rex_hsqc_lek_x00

with antiphase_flg set to 'y'

Journal of Biomolecular NMR (2011) 50, 13-8
"""


import numpy as np

from chemex.experiments.cpmg import cpmg_profile
from chemex.spindynamics import basis, default

EXP_DETAILS = {
    "carrier": {"type": float},
    "time_t2": {"type": float},
    "pw90": {"type": float},
    "time_equil": {"default": 0.0, "type": float},
}


class Profile(cpmg_profile.CPMGProfile):
    """TODO: class docstring."""

    def __init__(self, name, measurements, exp_details, model):
        super().__init__(name, measurements, exp_details, model)

        self.exp_details = self.check_exp_details(exp_details, expected=EXP_DETAILS)

        self.liouv = basis.Liouvillian(
            system="ixyzsz",
            state_nb=self.model.state_nb,
            atoms=self.peak.atoms,
            h_larmor_frq=self.conditions["h_larmor_frq"],
            equilibrium=False,
        )

        self.liouv.carrier_i = self.exp_details["carrier"]
        self.liouv.w1_i = 2.0 * np.pi / (4.0 * self.exp_details["pw90"])

        self.detect = self.liouv.detect["2izsz_a"]

        self.t_cps = {
            ncyc: self.exp_details["time_t2"] / (4.0 * ncyc) - self.exp_details["pw90"]
            for ncyc in self.ncycs[self.ncycs > 0]
        }

        self.t_neg = -2.0 * self.exp_details["pw90"] / np.pi

        self.delays = [self.t_neg, self.exp_details["time_equil"]]
        self.delays.extend(self.t_cps.values())

        self.map_names, self.params = default.create_params(
            basis=self.liouv,
            model=self.model,
            nuclei=self.peak.names,
            conditions=self.conditions,
            hn_ap_constraints=True,
        )

        for name, full_name in self.map_names.items():
            if name.startswith(("dw", "r2_i_a")):
                self.params[full_name].set(vary=True)

    def calculate_unscaled_profile(self, **parvals):
        """TODO: class docstring."""

        self.liouv.update(**parvals)

        # Calculation of all the needed propagators
        pulses = self.liouv.pulses_90_180_i()
        delays = dict(zip(self.delays, self.liouv.delays(self.delays)))

        p_180pmx = 0.5 * (pulses["180px"] + pulses["180mx"])  # +/- phase cycling

        # Simulate the CPMG block as function of ncyc
        mag0 = self.liouv.compute_mag_eq(term="2izsz", **parvals)

        profile = []

        for ncyc in self.ncycs:

            if ncyc == 0:

                mag = (
                    self.detect
                    @ delays[self.exp_details["time_equil"]]
                    @ pulses["90px"]
                    @ p_180pmx
                    @ pulses["90px"]
                    @ mag0
                )

            else:
                p_cp = np.linalg.matrix_power(
                    delays[self.t_cps[ncyc]]
                    @ pulses["180py"]
                    @ delays[self.t_cps[ncyc]],
                    int(ncyc),
                )
                mag = (
                    self.detect
                    @ delays[self.exp_details["time_equil"]]
                    @ pulses["90px"]
                    @ delays[self.t_neg]
                    @ p_cp
                    @ p_180pmx
                    @ p_cp
                    @ delays[self.t_neg]
                    @ pulses["90px"]
                    @ mag0
                )

            profile.append(np.float64(self.liouv.collapse(mag)))

        return np.asarray(profile)

    def ncycs_to_nu_cpmgs(self, ncycs=None):
        """Calculate the pulsing frequency, v(CPMG), from ncyc values."""
        if ncycs is None:
            ncycs = self.ncycs

        return ncycs / self.exp_details["time_t2"]



"""The parameters module contains the code for handling of the experimental and
fitting parameters."""
import ast
import configparser
import copy
import pathlib
import re
from difflib import get_close_matches

import lmfit
import numpy as np
from asteval import astutils

from chemex import peaks
from chemex import util

NAME_MARKERS = {
    "name": "__n_{}_n__",
    "nuclei": "__r_{}_r__",
    "temperature": "__t_{}_t__",
    "h_larmor_frq": "__b_{}_b__",
    "p_total": "__p_{}_p__",
    "l_total": "__l_{}_l__",
}

FRIENDLY_MARKERS = {
    "name": "{}",
    "nuclei": "NUC->{}",
    "temperature": "T->{:.1f}C",
    "h_larmor_frq": "B0->{:.1f}MHz",
    "p_total": "[P]->{:e}M",
    "l_total": "[L]->{:e}M",
}

RE_QUALIFIERS = re.compile(
    """
        (^\s*(?P<name>\w+)) |
        (NUC\s*->\s*(?P<nuclei>(\w|-)+)) |
        (T\s*->s*(?P<temperature>{0})) |
        (B0\s*->\s*(?P<h_larmor_frq>{0})) |
        (\[P\]\s*->\s*(?P<p_total>{0})) |
        (\[L\]\s*->\s*(?P<l_total>{0})) |
    """.format(
        "[-+]?[0-9]*\.?[0-9]+(e[-+]?[0-9]+)?"
    ),
    re.IGNORECASE | re.VERBOSE,
)

# Regular expression to pick values of the form: intial value [min, max]
RE_VALUE_MIN_MAX = re.compile(
    """
        ^\s*
        (?P<value>{0})?\s*
        (\[\s*(?P<min>({0}|{1}))\s*,\s*(?P<max>({0}|{1}))\s*\]\s*)?
        .*$
    """.format(
        "[-+]?[0-9]*\.?[0-9]+(e[-+]?[0-9]+)?", "[-+]?inf"
    ),
    re.IGNORECASE | re.VERBOSE,
)

# Regular expression to pick values of the form: intial value [min, max, brute_stepsize]
RE_VALUE_MIN_MAX_BRUTE = re.compile(
    """
        ^\s*
        (?P<value>{0})?\s*
        (\[\s*(?P<min>({0}|{1}))\s*,\s*(?P<max>({0}|{1}))\s*,\s*(?P<brute_step>({0}|None))\s*\]\s*)?
        .*$
    """.format(
        "[-+]?[0-9]*\.?[0-9]+(e[-+]?[0-9]+)?", "[-+]?inf"
    ),
    re.IGNORECASE | re.VERBOSE,
)

RE_PARNAME = re.compile(
    """
        (__n_(?P<name>\w+)_n__)?
        (__r_(?P<nuclei>(\w|-)+)_r__)?
        (__t_(?P<temperature>{0})_t__)?
        (__b_(?P<h_larmor_frq>{0})_b__)?
        (__p_(?P<p_total>{0})_p__)?
        (__l_(?P<l_total>{0})_l__)?
    """.format(
        "[-+]?[0-9]*\.?[0-9]+(e[-+]?[0-9]+)?"
    ),
    re.IGNORECASE | re.VERBOSE,
)


class MakeTranslate:
    """MakeTranslate class for translating parameter names.

    From: https://www.oreilly.com/library/view/python-cookbook-2nd/0596007973/ch01s19.html
    """

    def __init__(self, *args, **kwds):
        self.dictionary = dict(*args, **kwds)
        self.re_expr = self.make_re_expr()

    def make_re_expr(self):
        """TODO: method docstring."""
        return re.compile("|".join(map(re.escape, self.dictionary)), re.IGNORECASE)

    def one_xlat(self, match):
        """TODO: method docstring."""
        return self.dictionary[match.group(0)]

    def __call__(self, text):
        return self.re_expr.sub(self.one_xlat, text)


EXPAND = MakeTranslate({"-": "__minus__", "+": "__plus__", ".": "__point__"})
COMPRESS = MakeTranslate({"__minus__": "-", "__plus__": "+", "__point__": "."})


class ParameterName:
    """ParameterName class."""

    def __init__(
        self,
        name=None,
        nuclei=None,
        temperature=None,
        h_larmor_frq=None,
        p_total=None,
        l_total=None,
    ):
        self.name = self.nuclei = self.temperature = self.h_larmor_frq = None
        self.p_total = self.l_total = None

        if name is not None:
            self.name = name.lower()

        if nuclei is not None:
            self.nuclei = peaks.Peak(nuclei).assignment

        if temperature is not None:
            self.temperature = round(float(temperature), 1)

        if h_larmor_frq is not None:
            self.h_larmor_frq = round(float(h_larmor_frq), 1)

        if p_total is not None:
            self.p_total = float(p_total)

        if l_total is not None:
            self.l_total = float(l_total)

    @classmethod
    def from_full_name(cls, full_name=None):
        """TODO: method docstring."""
        if full_name is None:
            full_name = ""

        full_name = COMPRESS(full_name)

        match = re.match(RE_PARNAME, full_name)
        qualifiers = {}
        if match is not None:
            qualifiers.update(match.groupdict())

        return cls(**qualifiers)

    @classmethod
    def from_section(cls, section=None):
        """TODO: method docstring."""
        if section is None:
            section = ""
        qualifiers = re_to_dict(RE_QUALIFIERS, section)

        return cls(**qualifiers)

    def update_nuclei(self, nuclei=None):
        """TODO: method docstring."""
        if nuclei is not None:
            self.nuclei = peaks.Peak(nuclei).assignment

        return self

    def to_full_name(self):
        """TODO: method docstring."""
        name_components = []

        for attribute, value in vars(self).items():
            if value is not None:
                name_components.append(NAME_MARKERS[attribute].format(value))

        full_name = EXPAND("".join(name_components))

        return full_name

    def to_section_name(self, nuclei=False):
        """TODO: method docstring."""

        name_components = []

        for attribute, value in vars(self).items():
            if (attribute != "nuclei" or nuclei) and value is not None:
                name_components.append(FRIENDLY_MARKERS[attribute].format(value))

        section_name = ", ".join(name_components).upper()

        return section_name

    def to_re(self):
        """TODO: method docstring."""
        re_components = [NAME_MARKERS["name"].format(EXPAND(self.name))]

        if self.nuclei is not None:
            group_name = peaks.Peak(self.nuclei)._resonances["i"]["group"]
            if not group_name:
                all_res = "\D?[0-9]+[abd-gi-mopr-z]*"
            else:
                all_res = ""
            re_components.append(
                NAME_MARKERS["nuclei"].format("".join([all_res, EXPAND(self.nuclei)]))
            )
        else:
            re_components.append(".*")

        if self.temperature is not None:
            re_components.append(
                NAME_MARKERS["temperature"].format(EXPAND(str(self.temperature)))
            )
        elif re_components[-1] != ".*":
            re_components.append(".*")

        if self.h_larmor_frq is not None:
            re_components.append(
                NAME_MARKERS["h_larmor_frq"].format(EXPAND(str(self.h_larmor_frq)))
            )
        elif re_components[-1] != ".*":
            re_components.append(".*")

        if self.p_total is not None:
            re_components.append(
                NAME_MARKERS["p_total"].format(EXPAND(str(self.p_total)))
            )
        elif re_components[-1] != ".*":
            re_components.append(".*")

        if self.l_total is not None:
            re_components.append(
                NAME_MARKERS["l_total"].format(EXPAND(str(self.l_total)))
            )
        elif re_components[-1] != ".*":
            re_components.append(".*")

        re_to_match = re.compile("".join(re_components), re.IGNORECASE)

        return re_to_match

    def __repr__(self):
        return self.to_section_name(nuclei=True)

    def __lt__(self, other):
        tuple_self = (
            self.name,
            self.temperature,
            self.h_larmor_frq,
            self.p_total,
            self.l_total,
            peaks.Peak(self.nuclei),
        )

        tuple_other = (
            other.name,
            other.temperature,
            other.h_larmor_frq,
            other.p_total,
            other.l_total,
            peaks.Peak(other.nuclei),
        )

        return tuple_self < tuple_other

    def intersection(self, other):
        """TODO: method docstring."""
        name = temperature = h_larmor_frq = p_total = l_total = None

        if self.name == other.name:
            name = self.name

        nuclei = (
            peaks.Peak(self.nuclei).intersection(peaks.Peak(other.nuclei)).assignment
        )

        if not nuclei:
            nuclei = None

        if self.temperature == other.temperature:
            temperature = self.temperature

        if self.h_larmor_frq == other.h_larmor_frq:
            h_larmor_frq = self.h_larmor_frq

        if self.p_total == other.p_total:
            p_total = self.p_total

        if self.l_total == other.l_total:
            l_total = self.l_total

        return ParameterName(
            name=name,
            nuclei=nuclei,
            temperature=temperature,
            h_larmor_frq=h_larmor_frq,
            p_total=p_total,
            l_total=l_total,
        )


def create_params(data):
    """Create the array of fitting parameters."""
    params = lmfit.Parameters()

    for profile in data:
        for name, param in profile.params.items():
            if name in params:
                vary = params[name].vary
            else:
                vary = False
            param._delay_asteval = True
            params[name] = param
            params[name].vary = vary | param.vary

    for p in params.values():
        p._delay_asteval = False

    return params


def set_params_from_config_file(params, config_filename):
    """Read the parameter file and set initial values and optional bounds and brute step size."""

    print(f"File Name: {config_filename}", end="\n\n")

    config = util.read_cfg_file(config_filename)

    print("{:<45s} {:<30s}".format("Section", "Matches"))
    print("{:<45s} {:<30s}".format("-------", "-------"))

    for section in config.sections():
        if section.lower() in ("global", "default"):
            print("{:<45s}".format(f"[{section}]"))

            for key, value in config.items(section):
                name = ParameterName.from_section(key)
                if value.count(",") == 2:
                    default = re_to_dict(RE_VALUE_MIN_MAX_BRUTE, value)
                else:
                    default = re_to_dict(RE_VALUE_MIN_MAX, value)
                default = {key: np.float64(val) for key, val in default.items()}
                matches = set_params(params, name, **default)

                print("  {:<43s} {:<30d}".format(f"({key})", len(matches)))

        else:
            name = ParameterName.from_section(section)

            pairs = []

            for key, value in config.items(section):
                if "file" in key:
                    for filename in value.split():
                        filename_ = pathlib.Path(filename)
                        filename_ = util.normalize_path(
                            config_filename.parent, filename_
                        )
                        pairs.extend(get_pairs_from_file(filename_, name))

                elif peaks.RE_PEAK_NAME.match(key):
                    name.update_nuclei(key)
                    pairs.append((copy.deepcopy(name), value))

            total_matches = set()

            for name, value in pairs:
                if value.count(",") == 2:
                    default = re_to_dict(RE_VALUE_MIN_MAX_BRUTE, value)
                else:
                    default = re_to_dict(RE_VALUE_MIN_MAX, value)
                default = {key: np.float64(val) for key, val in default.items()}
                matches = set_params(params, name, **default)
                total_matches.update(matches)

            print("{:<45s} {:<30d}".format(f"[{section}]", len(total_matches)))


def get_pairs_from_file(filename, name):
    """Read residue specific values for fitting parameters from a file.

    The file should be formatted like a Sparky peak list.
    Examples:
      * To set G23N to 105.0 and G23H to 8.0:
          G23N-H  105.0  8.0
      * To set a parameter depending on multiple nuclei (e.g., G23N and G23H):
          G23N-H  -93.0

    """
    pairs = []

    with open(filename) as f:
        for line in f:
            if "Assignment" in line:
                continue

            line = remove_comments(line, "#;")
            line = re.sub("\s*\[\s*", "[", line)
            line = re.sub("\s*\]\s*", "]", line)
            elements = line.split()

            if len(elements) > 1:
                peak = peaks.Peak(elements[0])
                n_resonances = len(peak)
                n_cols = len(elements[1:])

                if n_cols == n_resonances:
                    for nuc_name, value in zip(peak.names.values(), elements[1:]):
                        name.update_nuclei(nuc_name)
                        pairs.append((copy.deepcopy(name), value))

                else:
                    name.update_nuclei(peak.assignment)
                    pairs.append((copy.deepcopy(name), elements[1]))

    return pairs


def set_param_status(params, items):
    """Set whether or not to vary a fitting parameter or to use a mathemetical expression."""

    vary = {"fix": False, "fit": True}

    for key, status in items:
        name = ParameterName.from_section(key)

        if status in vary:
            set_params(params, name, vary=vary[status], expr="")
        else:
            set_param_expr(params, name, expr=status)


def set_param_expr(params, name, expr=None):
    """Set an optional parameter expression, used to constrain its value during
    the fit."""

    if expr is None:
        expr = ""

    if not isinstance(name, ParameterName):
        name = ParameterName.from_section(name)

    names_full = [name_full for name_full in params if name.to_re().match(name_full)]
    names_expr = astutils.get_ast_names(ast.parse(expr))
    names_full_expr = {
        name: [
            name_full_expr
            for name_full_expr in params
            if ParameterName.from_section(name).to_re().match(name_full_expr)
        ]
        for name in names_expr
    }

    matches = set()

    for name_full in names_full:

        expr_ = expr

        for name_expr in names_expr:
            name_full_expr = get_close_matches(
                name_full, names_full_expr[name_expr], n=1
            )[0]
            expr_ = expr_.replace(name_expr, name_full_expr)

        params[name_full].expr = expr_

        matches.add(name_full)

    return matches


def set_params(
    params,
    name_short,
    value=None,
    vary=None,
    min=None,
    max=None,
    expr=None,
    brute_step=None,
):
    """Set the initial value and (optional) bounds and brute step size for parameters."""
    matches = set()
    name_short_re = name_short.to_re()

    for name, param in params.items():
        if name_short_re.match(name):
            if expr is None and param.expr and vary is None:
                param.value = value
            else:
                param.set(value, vary, min, max, expr, brute_step)
            matches.add(name)

    return matches


def write_par(params, path):
    """Write the fitting parameters and their uncertainties to a file."""
    filename = path / "parameters.fit"

    print(f"  * {filename}")

    par_dict = {}

    for name, param in params.items():

        par_name = ParameterName.from_full_name(name)

        if par_name.nuclei is None:  # global parameter
            name_print = par_name
            section = "GLOBAL"

        else:  # residue-specific parameter
            name_print = peaks.Peak(par_name.nuclei)
            section = par_name.to_section_name()

        if not param.vary and param.expr is None:
            val_print = f"{param.value: .5e} ; fixed"
        elif param.stderr is None:
            val_print = f"{param.value: .5e}  ; error not calculated"
        elif param.expr:
            val_print = "{: .5e} +/- {:.5e} ; constrained".format(
                param.value, param.stderr
            )
        else:
            val_print = f"{param.value: .5e} +/- {param.stderr:.5e}"

        par_dict.setdefault(section, []).append((name_print, val_print))

    cfg = configparser.ConfigParser()
    cfg.optionxform = str

    section_global = par_dict.pop("GLOBAL", None)

    if section_global is not None:
        cfg.add_section("GLOBAL")
        for name, val in sorted(section_global):
            cfg.set("GLOBAL", str(name), val)

    for section, name_vals in sorted(par_dict.items()):
        cfg.add_section(section)
        for peak, val in sorted(name_vals):
            cfg.set(section, peak.assignment.upper(), val)

    with open(filename, "w") as f:
        cfg.write(f)


def write_constraints(params, path):
    """Write the (optional) parameter expression constraints to a file."""
    filename = path / "constraints.fit"

    print(f"  * {filename}")

    param_dict = dict()

    for name, param in params.items():

        par_name = ParameterName.from_full_name(name)

        if param.expr:
            name_formatted = "[{}]".format(par_name.to_section_name(nuclei=True))
            expr_formatted = param.expr
            for name_dep in astutils.get_ast_names(ast.parse(param.expr)):
                par_name_dep = ParameterName.from_full_name(name_dep)
                if str(par_name_dep):
                    expr_formatted = expr_formatted.replace(
                        name_dep,
                        "[{}]".format(par_name_dep.to_section_name(nuclei=True)),
                    )

            param_dict[par_name] = f"{name_formatted} = {expr_formatted}\n"

    with open(filename, "w") as f:
        for name, constraint in sorted(param_dict.items()):
            f.write(constraint)


def remove_comments(line, sep):
    """Remove (optional) comments."""
    for a_sep in sep:
        line = line.split(a_sep)[0]

    return line.strip()


def re_to_dict(re_to_match, text):
    """TODO: function docstring."""
    return {
        match_key: match_value
        for match in re_to_match.finditer(text)
        for match_key, match_value in match.groupdict().items()
        if match_value is not None
    }



"""1H-13C(methyl) - Multiple Quantum CPMG (2-state)

Analyzes HyCx methyl group multiple quantum CPMG measured on site-specific
13CH3-labeled methyl groups in a highly deuterated background.  This is a
simplified basis set, which assumes you are on-resonance for 13C (ie, off-
resonance effects are not taken into account) as described in the reference:

[HxCx(a), HyCx(a), HxCy(a), HyCy(a),
 HxCx(b), HyCx(b), HxCy(b), HyCy(b)]

Note
----
This calculation is designed specifically to analyze data from the experiment
found in the reference and can be run with either small_protein_flag='y' or 'n'.

Lewis Kay experiment: hmqc_CH3_exchange_bigprotein_*00_lek_v2

Reference
---------

Journal of the American Chemical Society (2004), 126, 3964-73

"""


import numpy as np

from chemex.experiments.cpmg import cpmg_profile
from chemex.spindynamics import basis, default

EXP_DETAILS = {
    "time_t2": {"type": float},
    "small_protein_flg": {"type": str, "default": "n"},
}


class Profile(cpmg_profile.CPMGProfile):
    """TODO: class docstring."""

    def __init__(self, name, measurements, exp_details, model):
        super().__init__(name, measurements, exp_details, model)

        self.exp_details = self.check_exp_details(exp_details, expected=EXP_DETAILS)

        self.liouv = basis.Liouvillian(
            system="ixysxy",
            state_nb=self.model.state_nb,
            atoms=self.peak.atoms,
            h_larmor_frq=self.conditions["h_larmor_frq"],
            equilibrium=False,
        )

        self.t_zeta = 1.0 / (8.0 * 125.3)
        self.t_cps = {
            ncyc: self.exp_details["time_t2"] / (4.0 * ncyc)
            for ncyc in self.ncycs[self.ncycs > 0]
        }

        self.delays = [self.t_zeta]
        self.delays.extend(self.t_cps.values())

        self.detect = self.liouv.detect["2iysx_a"]

        self.small_protein_flg = (
            self.exp_details["small_protein_flg"].lower().startswith("y")
        )

        self.map_names, self.params = default.create_params(
            basis=self.liouv,
            model=self.model,
            nuclei=self.peak.names,
            conditions=self.conditions,
        )

        for name, full_name in self.map_names.items():
            if name.startswith(("dw", "r2_mq_a")):
                self.params[full_name].set(vary=True)

    def calculate_unscaled_profile(self, **parvals):
        """TODO: class docstring."""

        self.liouv.update(**parvals)

        # Calculation of all the needed propagators
        delays = dict(zip(self.delays, self.liouv.delays(self.delays)))

        # Simulate the CPMG block as function of ncyc
        mag0 = self.liouv.compute_mag_eq(term="2iysx", **parvals)

        profile = []

        if self.small_protein_flg:
            mag0 = (
                delays[self.t_zeta]
                @ self.liouv.perfect180["sx"]
                @ self.liouv.perfect180["ix"]
                @ delays[self.t_zeta]
                @ mag0
            )

        for ncyc in self.ncycs:

            if ncyc == 0:

                mag = self.detect @ mag0

            else:

                p_cpy = np.linalg.matrix_power(
                    delays[self.t_cps[ncyc]]
                    @ self.liouv.perfect180["iy"]
                    @ delays[self.t_cps[ncyc]],
                    int(ncyc),
                )

                mag = self.detect @ p_cpy @ self.liouv.perfect180["sx"] @ p_cpy @ mag0

            profile.append(np.float64(self.liouv.collapse(mag)))

        return np.asarray(profile)

    def ncycs_to_nu_cpmgs(self, ncycs=None):
        """Calculate the pulsing frequency, v(CPMG), from ncyc values."""

        if ncycs is None:
            ncycs = self.ncycs

        return ncycs / self.exp_details["time_t2"]



"""TODO: module docstring."""

import abc

import numpy as np

from chemex.spindynamics import util

EXP_DETAILS = {"name": {"type": str}}

CONDITIONS = {
    "h_larmor_frq": {"type": float},
    "temperature": {"type": float},
    "p_total": {"default": None, "type": float},
    "l_total": {"default": None, "type": float},
}


class BaseProfile(metaclass=abc.ABCMeta):
    """TODO: class docstring."""

    BOOLEAN_STATES = {
        "1": True,
        "yes": True,
        "true": True,
        "on": True,
        "0": False,
        "no": False,
        "false": False,
        "off": False,
    }

    def __init__(self, name=None, exp_details=None, model=None):
        """TODO: method docstring."""

        if name is None:
            name = ""

        self.profile_name = name
        self.model = util.parse_model(model)
        self.experiment_name = self.check_exp_details(exp_details, EXP_DETAILS)["name"]
        self.conditions = self.check_exp_details(exp_details, CONDITIONS)
        self.val = None
        self.err = None
        self.data = None
        self.mask = None

    def __len__(self):
        return self.val.size

    @abc.abstractmethod
    def print_profile(self, params=None):
        """TODO: method docstring."""
        pass

    @abc.abstractmethod
    def filter_points(self, params=None):
        """TODO: method docstring."""
        pass

    @abc.abstractmethod
    def calculate_residuals(self, params):
        """Calculate the residuals between the experimental and back-calculated
        values."""
        pass

    @staticmethod
    def check_exp_details(exp_details=None, expected=None):
        """Check for experimental parameters and convert them to their appropriate
        type."""

        if expected is None:
            expected = {}

        details = {}
        missing = []

        for name, item in expected.items():
            default = item.get("default")
            value = exp_details.get(name, default)
            required = "default" not in item
            if value is None and required:
                missing.append("'" + name + "'")
            elif isinstance(value, str):
                value = value.split()
            if value is None:
                details[name] = None
            else:
                details[name] = np.array(value, dtype=item.get("type")).reshape(-1)

                if details[name].shape == (1,):
                    details[name] = details[name][0]

        if missing:
            exit(
                "\n  - Missing experimental parameter detected. Please set: "
                "{:s}".format(", ".join(missing))
            )

        return details

    def get_bool(self, value):

        if value.lower() not in self.BOOLEAN_STATES:
            raise ValueError("Not a boolean: %s" % value)

        return self.BOOLEAN_STATES[value.lower()]



"""Utility functions for bases."""
import collections
import re

import numpy as np
from scipy import linalg

SIGN = np.array([1.0, -1.0])


def compute_propagator_standard(liouvillian, time):
    """TODO: function docstring."""

    return linalg.expm(liouvillian * time)


def compute_propagator_beff(liouvillian, time):
    """TODO: function docstring."""

    s, vr = linalg.eig(liouvillian)
    vri = linalg.inv(vr)

    sl = np.where(abs(s.imag) < 1e-6)[0]

    vri = vri[sl, :]
    diag = np.diag(np.exp(s[sl] * time))
    vr = vr[:, sl]

    return vr.dot(diag).dot(vri).real


def compute_propagators_from_time_series(liouvillian, times):
    """TODO: function docstring."""

    s, vr = linalg.eig(liouvillian)
    vri = linalg.inv(vr)

    propagators = {
        t: vr.dot(np.diag(np.exp(s * t))).dot(vri).real
        for t in set(times)
        if abs(t) != np.inf
    }

    return propagators


def calculate_shift_ex_2st(pb=0.0, kex_ab=0.0, domega_i_ab=0.0, r2_i_a=0.0, r2_i_b=0.0):
    """Correct major and minor peak positions in presence of exchange."""
    pa = 1.0 - pb
    kab, kba = kex_ab * np.asarray([pb, pa])

    k2ab = r2_i_a + kab
    k2ba = r2_i_b + kba - 1j * domega_i_ab

    k2ex = k2ab + k2ba
    fac = ((k2ab - k2ba) ** 2 + 4.0 * kab * kba) ** 0.5

    nu1 = (0.5 * (-k2ex + fac)).imag
    nu2 = (0.5 * (-k2ex - fac)).imag

    condition = abs(nu1) < abs(nu2)

    nu1_ = np.where(condition, nu1, nu2)
    nu2_ = np.where(condition, nu2, nu1)

    return nu1_, nu2_


def correct_intensities(
    magz_a=1.0, magz_b=0.0, pb=0.0, kex=0.0, dw=0.0, r2_i=0.0, dr2_i=0.0
):
    """Correct major and minor peak intensities in presence of exchange."""
    kab = kex * pb
    kba = kex - kab

    k2ab = r2_i + kab
    k2ba = r2_i + dr2_i - 1j * dw + kba

    k2ex = k2ab + k2ba
    fac = ((k2ab - k2ba) ** 2 + 4.0 * kab * kba) ** 0.5

    nu1, nu2 = 0.5 * (-k2ex + SIGN * fac)

    magz_a_c = +((kab - nu2 - k2ab) * magz_a + (kba + nu1 + k2ab) * magz_b) / (
        nu1 - nu2
    )

    magz_b_c = -((kab - nu1 - k2ab) * magz_a + (kba + nu2 + k2ab) * magz_b) / (
        nu1 - nu2
    )

    if abs(nu1.imag) > abs(nu2.imag):
        magz_a_c, magz_b_c = magz_b_c, magz_a_c

    if abs(np.angle(magz_a_c)) <= 0.5 * np.pi:
        magz_a_c = +abs(magz_a_c)
    else:
        magz_a_c = -abs(magz_a_c)

    if abs(np.angle(magz_b_c)) <= 0.5 * np.pi:
        magz_b_c = +abs(magz_b_c)
    else:
        magz_b_c = -abs(magz_b_c)

    return magz_a_c, magz_b_c


def parse_model(name):
    Model = collections.namedtuple("Model", ["name", "state_nb", "kind"])
    match = re.match("(\d)st\.(\w+)", name, re.IGNORECASE)
    if match:
        state_nb = int(match.group(1))
        kind = match.group(2)
        return Model(name, state_nb, kind)
    else:
        raise NameError(f"The model name '{name}' is not correct.")



"""Read the experimental CEST data."""

import importlib

import numpy as np
from scipy import interpolate, linalg, signal, stats


def read_profiles(path, filenames, details, model, included=None, excluded=None):
    """Read the CEST profiles."""

    experiment_type = details["type"].split(".")
    details["name"] = name_experiment(details)
    experiment_module = importlib.import_module(
        ".".join(
            ["chemex.experiments", experiment_type[0], "profiles", experiment_type[1]]
        )
    )

    error = details.get("error", "file")
    if error not in {"file", "auto"}:
        print("Warning: The 'error' option should either be 'file' or ")
        print("'auto'. Using the default 'file' option.")
        error = "file"

    error_values = []

    if included is None:
        included = filenames.keys()
    else:
        included = [_.lower() for _ in included]

    if excluded is None:
        excluded = []
    else:
        excluded = [_.lower() for _ in excluded]

    Profile = getattr(experiment_module, "Profile")

    dtype = [("b1_offsets", "<f8"), ("intensities", "<f8"), ("intensities_err", "<f8")]

    profiles = []

    for profile_name, filename in filenames.items():

        full_path = path / filename

        measurements = np.loadtxt(full_path, dtype=dtype)
        measurements.sort(order="b1_offsets")

        profile = Profile(profile_name, measurements, details, model)

        is_included = profile_name.lower() in included
        is_not_excluded = profile_name.lower() not in excluded

        if is_included and is_not_excluded:
            profiles.append(profile)

        if error == "auto":
            excl_reference = np.logical_not(profile.reference)
            error_values.append(estimate_noise(profile.val[excl_reference]))

    error = details.get("error", "file")

    if error == "auto":

        error_value = np.mean(error_values)

        for profile in profiles:
            profile.err[:] = error_value

    ndata = sum(len(profile.val) for profile in profiles)

    return profiles, ndata


def name_experiment(experiment_details=None):
    """Generate a unique name for the experiment."""
    if experiment_details is None:
        experiment_details = dict()

    if "name" in experiment_details:
        name = experiment_details["name"].strip().replace(" ", "_")

    else:
        exp_type = experiment_details["type"].replace(".", "_")
        h_larmor_frq = float(experiment_details["h_larmor_frq"])
        temperature = float(experiment_details["temperature"])
        b1_frq = float(experiment_details["b1_frq"])
        time_t1 = float(experiment_details["time_t1"])

        name = "{:s}_{:.0f}Hz_{:.0f}ms_{:.0f}MHz_{:.0f}C".format(
            exp_type, b1_frq, time_t1 * 1e3, h_larmor_frq, temperature
        ).lower()

    return name


def estimate_noise(values):
    """Estimate the uncertainty in the CEST profile.

    # TODO: add reference to this method...

    """
    size = len(values)

    fda = [
        [1, -1],
        [1, -2, 1],
        [1, -3, 3, -1],
        [1, -4, 6, -4, 1],
        [1, -5, 10, -10, 5, -1],
        [1, -6, 15, -20, 15, -6, 1],
    ]

    fda = [np.array(a_fda) / linalg.norm(a_fda) for a_fda in fda]

    percents = np.array([0.05] + list(np.arange(0.1, 0.40, 0.025)))
    percent_points = stats.norm.ppf(1.0 - percents)

    sigma_est = []

    for fdai in fda:
        noisedata = signal.convolve(values, fdai, mode="valid")
        ntrim = len(noisedata)

        if ntrim >= 2:
            noisedata.sort()

            xaxis = 0.5 + np.arange(1, ntrim + 1)
            xaxis /= ntrim + 0.5

            sigmas = []

            f_interpalated = interpolate.interp1d(xaxis, noisedata, "linear")

            for a_perc, a_z in zip(percents, percent_points):
                try:
                    val = (f_interpalated(1.0 - a_perc) - f_interpalated(a_perc)) / (
                        2.0 * a_z
                    )
                    sigmas.append(val)
                except ValueError:
                    pass

            sigma_est.append(np.median(sigmas))

    noisevar = np.median(sigma_est) ** 2
    noisevar /= 1.0 + 15.0 * (size + 1.225) ** -1.245

    return np.sqrt(noisevar)



"""The parsing module contains the code for the parsing of command-line
arguments."""
import argparse
import importlib
import pathlib
import pkgutil
import sys

from chemex import _version
from chemex import chemex
from chemex import experiments
from chemex import util
from chemex.tools import pick_cest
from chemex.tools import plot_param

FITMETHODS = {
    "cobyla",
    "slsqp",
    "tnc",
    "nelder",
    "cg",
    "bfgs",
    "powell",
    "l-bfgsb",
    "least_squares",
    "lbfgsb",
    "differential_evolution",
    "leastsq",
    "brute",
    "basinhopping",
    "ampgo",
}


class MyParser(argparse.ArgumentParser):
    """Subclass of ArgumentParser to override the error method."""

    def error(self, message):
        sys.stderr.write(f"error: {message}\n\n")
        self.print_help()
        sys.exit(2)


def build_parser():
    """Parse the command-line arguments."""
    description = (
        "ChemEx is an analysis program for chemical exchange detected by "
        "NMR. It is designed to take almost any kind of NMR data to aid the "
        "analysis, but the principle techniques are CPMG relaxation "
        "dispersion and Chemical Exchange Saturation Transfer."
    )

    parser = MyParser(description=description, prog="chemex")

    parser.add_argument(
        "--version", action="version", version=f"{parser.prog} {_version.version}"
    )

    commands = parser.add_subparsers(dest="commands")

    # parser for the positional argument "info"
    info_parser = commands.add_parser(
        "info",
        help="Shows experiments that can be fit",
        description="Enter the name of an experiment to obtain more info about it.",
    )

    info_parser.set_defaults(func=get_info)

    experiments_parser = info_parser.add_subparsers(dest="experiments")
    experiments_parser.required = True

    experiments_names = list_experiments()

    for name in experiments_names:
        experiment_module = importlib.import_module(name)
        experiment_shortname = compress_name(name)

        experiments_parser.add_parser(
            experiment_shortname,
            help=get_description_from_module(experiment_module),
            add_help=False,
        )

    # parser for the positional argument "fit"
    fit_parser = commands.add_parser("fit", help="Starts a fit", prefix_chars="+-")

    fit_parser.set_defaults(func=chemex.fit)

    fit_parser.add_argument(
        "-e",
        dest="experiments",
        type=pathlib.Path,
        metavar="FILE",
        nargs="+",
        required=True,
        help="Input files containing experimental setup and data location",
    )

    fit_parser.add_argument(
        "-d",
        dest="model",
        metavar="MODEL",
        default="2st.pb_kex",
        help="Exchange model used to fit the data",
    )

    fit_parser.add_argument(
        "-p",
        dest="parameters",
        type=pathlib.Path,
        metavar="FILE",
        nargs="+",
        required=True,
        help="Input file containing the initial values of fitting parameters",
    )

    fit_parser.add_argument(
        "-m",
        dest="method",
        type=pathlib.Path,
        metavar="FILE",
        help="Input file containing the fitting method",
    )

    fit_parser.add_argument(
        "-o",
        dest="out_dir",
        type=pathlib.Path,
        metavar="DIR",
        default="./Output",
        help="Directory for output files",
    )

    fit_parser.add_argument(
        "--noplot", action="store_true", help="No plots of the fits"
    )

    fit_parser.add_argument(
        "-f",
        dest="fitmethod",
        metavar="FITMETHOD",
        default="leastsq",
        choices=sorted(FITMETHODS),
        help="Specify the fitting method",
    )

    selection = fit_parser.add_mutually_exclusive_group()
    selection.add_argument(
        "+r",
        dest="res_incl",
        metavar="ID",
        nargs="+",
        help="residue(s) to include in the fit",
    )
    selection.add_argument(
        "-r",
        dest="res_excl",
        metavar="ID",
        nargs="+",
        help="residue(s) to exclude from the fit",
    )

    simulation = fit_parser.add_mutually_exclusive_group()
    simulation.add_argument(
        "--mc", metavar="N", type=int, help="Run N Monte-Carlo simulations"
    )
    simulation.add_argument(
        "--bs", metavar="N", type=int, help="Run N Bootstrap simulations"
    )

    # parser for the positional argument "pick_cest"
    pick_cest_parser = commands.add_parser(
        "pick_cest", help="Plot CEST profiles for dip picking"
    )

    pick_cest_parser.set_defaults(func=pick_cest.pick_cest)

    pick_cest_parser.add_argument(
        "-e",
        dest="experiments",
        type=pathlib.Path,
        metavar="FILE",
        nargs=1,
        required=True,
        help="Input files containing experimental setup and data location",
    )

    pick_cest_parser.add_argument(
        "-o",
        dest="out_dir",
        type=pathlib.Path,
        metavar="DIR",
        default="./Output",
        help="Directory for output files",
    )

    # parser for the positional argument "pick_cest"
    plot_param_parser = commands.add_parser(
        "plot_param", help="Plot one selected parameter from a 'parameters.fit' file"
    )

    plot_param_parser.set_defaults(func=plot_param.plot_param)

    plot_param_parser.add_argument(
        "-p",
        dest="parameters",
        type=pathlib.Path,
        metavar="FILE",
        required=True,
        help="Input file containing the initial values of fitting parameters",
    )

    plot_param_parser.add_argument(
        "-n",
        dest="parname",
        metavar="NAME",
        required=True,
        help="Name of the parameter to plot",
    )

    return parser


def get_description_from_module(module):
    return module.__doc__.strip().splitlines()[0]


def list_experiments():
    path = experiments.__path__
    prefix = experiments.__name__ + "."
    names = []
    types_modules = pkgutil.walk_packages(path, prefix)
    for __, name, ispkg in types_modules:
        module = importlib.import_module(name)
        if not ispkg and hasattr(module, "Profile"):
            names.append(name)
    return names


def get_info(args):
    name = expand_name(args.experiments)
    module = importlib.import_module(name)
    util.header1('Description of the "' "{}" '" experiment'.format(args.experiments))
    print(module.__doc__)


def compress_name(name):
    tree = name.split(".")
    shortname = ".".join((tree[-3], tree[-1]))
    return shortname


def expand_name(name):
    type1, type2 = name.split(".")
    fullname = ".".join(["chemex.experiments", type1, "profiles", type2])
    return fullname



"""The chemex module provides the entry point for the chemex script."""
import copy
import shutil

import numpy as np

from chemex import _version
from chemex import cli
from chemex import datasets
from chemex import fitting
from chemex import parameters
from chemex import util

LOGO = r"""
* * * * * * * * * * * * * * * * * * * * * * * * *
*      ________                   ______        *
*     / ____/ /_  ___  ____ ___  / ____/  __    *
*    / /   / __ \/ _ \/ __ `__ \/ __/ | |/_/    *
*   / /___/ / / /  __/ / / / / / /____>  <      *
*   \____/_/ /_/\___/_/ /_/ /_/_____/_/|_|      *
*                                               *
*   Analysis of NMR Chemical Exchange data      *
*                                               *
*   Version: {:<34s} *
*                                               *
* * * * * * * * * * * * * * * * * * * * * * * * *
""".format(
    _version.version
)


def main():
    """Do all the magic."""
    print(LOGO)

    parser = cli.build_parser()
    args = parser.parse_args()

    if args.commands is None:
        parser.print_help()
    else:
        args.func(args)


def fit(args):
    # Read experimental setup and data
    data = datasets.read_data(args)

    # Create and update initial values of fitting/fixed parameters
    util.header1("Reading Default Parameters")
    params = parameters.create_params(data)

    for name in args.parameters:
        parameters.set_params_from_config_file(params, name)

    # Filter datapoints out if necessary (e.g., on-resonance filter CEST)
    for profile in data:
        profile.filter_points(params)

    data.ndata = sum([len(profile.val) for profile in data])

    # Customize the output directory
    output_dir = args.out_dir

    if args.res_incl and len(args.res_incl) == 1:
        output_dir = output_dir / args.res_incl.pop().upper()

    result = fit_write_plot(args, params, data, output_dir)

    if args.bs or args.mc:
        if args.bs:
            nmb = args.bs
        else:
            nmb = args.mc

        formatter_output_dir = "".join(["{:0", str(int(np.log10(nmb)) + 1), "d}"])

        for index in range(1, nmb + 1):
            if args.bs:
                data_index = data.make_bs_dataset()
            else:
                data_index = data.make_mc_dataset(result.params)

            output_dir_ = output_dir / formatter_output_dir.format(index)

            params_mc = copy.deepcopy(result.params)

            fit_write_plot(args, params_mc, data_index, output_dir_)


def fit_write_plot(args, params, data, output_dir):
    """Perform the fit, write the output files and plot the results."""

    result = fitting.run_fit(args.method, params, data, args.fitmethod)

    output_dir.mkdir(parents=True, exist_ok=True)

    write_results(result, data, args.method, output_dir)

    if not args.noplot:
        plot_results(result, data, output_dir)

    return result


def write_results(result, data, method, output_dir):
    """Write the results of the fit to output files.

    The files below are created and contain the following information:
      - parameters.fit: fitting parameters and their uncertainties
      - contstraints.fit: expression used for constraining parameters
      - *.dat: experimental and fitted data
      - statistics.fit: statistics for the fit

    """
    util.header1("Writing Results")

    print("\nFile(s):")

    if method:
        shutil.copyfile(method, output_dir / "fitting-method.cfg")

    parameters.write_par(result.params, path=output_dir)
    parameters.write_constraints(result.params, path=output_dir)
    data.write_to(result.params, path=output_dir)
    fitting.write_statistics(result, path=output_dir)


def plot_results(result, data, path):
    """Plot the experimental and fitted data."""
    from chemex.experiments import plotting

    util.header1("Plotting Data")

    print("\nFile(s):")

    path_plots = path / "Plots"
    path_plots.mkdir(parents=True, exist_ok=True)

    try:
        plotting.plot_data(data, result.params, path=path_plots)
    except KeyboardInterrupt:
        print(" - Plotting cancelled")

    if result.method == "brute":
        labels = [
            parameters.ParameterName.from_full_name(var).name.upper()
            for var in result.var_names
        ]
        outfile = path / "results_brute.pdf"
        plotting.plot_results_brute(result, varlabels=labels, output=outfile)
        print((f"  * {outfile}"))



import contextlib
import sys

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Button
from scipy import interpolate

from chemex import datasets
from chemex import peaks


def pick_cest(args):

    if len(args.experiments) > 1:
        sys.exit(
            "\nError: Multiple experiment files were given. 'chemex pick_cest' "
            "should only be run with a single experiment files.\n"
        )

    # Read experimental setup and data
    data = datasets.read_data(args)

    if not data.data[0].experiment_name.startswith("cest"):
        sys.exit(
            "\nError: The command 'chemex pick_cest' only works with CEST "
            "experiments.\n"
        )

    callback = Buttons(data, args.out_dir)

    axprevious = plt.axes([0.825, 0.1, 0.075, 0.075])
    axnext = plt.axes([0.9, 0.1, 0.075, 0.075])

    bprevious = Button(axprevious, "Previous")
    bprevious.on_clicked(callback.previous)

    bnext = Button(axnext, "Next")
    bnext.on_clicked(callback.next)

    plt.show()


class Buttons:

    KWARGS = {"color": "0.74", "linewidth": 1.0}
    LSTYLE = {"a": "-", "b": ":"}
    TEXT_Y = {"a": 0.8, "b": 0.75}

    def __init__(self, data, path):

        self.data = data
        self.out = path

        self.names = sorted(
            [peaks.Peak(profile.profile_name).names["i"] for profile in self.data]
        )

        self.curve, self.curve_sp = None, None
        self.cs_a, self.cs_b = {}, {}
        self.lines = []
        self.cid = None

        self.fig, self.axis = plt.subplots(figsize=(12, 5))
        self.fig.subplots_adjust(left=0.07, bottom=0.1, right=0.8, top=0.9)

        self.index = -1
        self.next(None)

    def previous(self, event):
        self._shift(-1)

    def next(self, event):
        self._shift(+1)

    def _shift(self, step):
        self.index += step
        self.index %= len(self.data)
        self._profile = self.data[self.index]
        self.name = peaks.Peak(self._profile.profile_name)
        self._profile_to_curve()
        self._clear_axis()
        self._plot()

    def _plot(self, event=None):
        self.cid = self.fig.canvas.mpl_connect("button_press_event", self._plot_lines)
        self._clear_axis()
        self._plot_lines()
        self._plot_profile()
        self.fig.canvas.draw()

    def _plot_profile(self):

        x, y = self.curve["x"], self.curve["y"]
        x_sp, y_sp = self.curve_sp["x"], self.curve_sp["y"]

        self.axis.plot(x, y, ".", color="C0")
        self.axis.plot(x_sp, y_sp, "--", color="C0", alpha=0.66)

        self._show_labels()

        self.fig.canvas.draw()

    def _show_labels(self):
        self.axis.set_title(str(self.name).upper())
        self.axis.set_xlabel(r"$^{15}$N (ppm)")
        self.axis.set_ylabel("$I/I_0$")

    def _profile_to_curve(self):
        profile = self._profile

        ref = profile.reference
        val_ref = np.mean(profile.val[ref])

        offsets = profile.b1_offsets_to_ppm()[~ref]
        vals = profile.val[~ref] / val_ref
        errs = profile.err[~ref] / abs(val_ref)

        spline = interpolate.CubicSpline(offsets, vals)
        fine_offsets = np.linspace(min(offsets), max(offsets), 1000)

        self.curve = {"x": offsets, "y": vals, "e": errs}
        self.curve_sp = {"x": fine_offsets, "y": spline(fine_offsets)}

    def _plot_lines(self, event=None):

        name = self.name.names["i"]

        cs_a = self.cs_a.get(name)
        cs_b = self.cs_b.get(name)

        xdata = self._get_click_position(event)

        if xdata is not None:
            if cs_a is None or cs_b is not None:
                cs_a = xdata
                cs_b = None
            else:
                cs_b = xdata

        self._clear_lines()

        if cs_a is not None:
            self._add_line(cs_a, "a")

        if cs_b is not None:
            dw_ab = cs_b - cs_a
            self._add_line(cs_b, "b")
            self._add_text_dw(dw_ab)

        self.cs_a[name] = cs_a
        self.cs_b[name] = cs_b

        self._save()

        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def _get_click_position(self, event=None):
        result = None
        if event and event.inaxes == self.axis:
            result = event.xdata
        return result

    def _add_line(self, position, state):
        text_ = fr"$\varpi_{state}$ = {position:.3f} ppm"
        text = self.fig.text(0.82, self.TEXT_Y[state], text_)
        line = self.axis.axvline(position, linestyle=self.LSTYLE[state], **self.KWARGS)
        self.lines.extend([line, text])

    def _add_text_dw(self, dw_ab):
        text_ = fr"$\Delta\varpi_{{ab}}$ = {dw_ab:.3f} ppm"
        text = self.fig.text(0.82, 0.7, text_)
        self.lines.append(text)

    def _clear_axis(self):
        self._clear_lines()
        self.axis.clear()

    def _clear_lines(self):
        while self.lines:
            self.lines.pop().remove()

    def _save(self):

        self.out.mkdir(parents=True, exist_ok=True)

        fname1 = self.out / "cs_a.txt"
        fname2 = self.out / "cs_b.txt"
        fname3 = self.out / "dw_ab.txt"

        with contextlib.ExitStack() as stack:

            file1 = stack.enter_context(fname1.open("w"))
            file2 = stack.enter_context(fname2.open("w"))
            file3 = stack.enter_context(fname3.open("w"))

            for name in self.names:

                cs_a = self.cs_a.get(name)
                cs_b = self.cs_b.get(name)

                if cs_a is None:
                    continue
                elif cs_b is None:
                    cs_b = cs_a

                dw_ab = cs_b - cs_a

                file1.write("{:10s} {:8.3f}\n".format(name.upper(), cs_a))
                file2.write("{:10s} {:8.3f}\n".format(name.upper(), cs_b))
                file3.write("{:10s} {:8.3f}\n".format(name.upper(), dw_ab))



import argparse
import configparser
import pathlib

from matplotlib import pyplot as plt

from chemex import parameters, peaks


def plot_param(args):

    params = configparser.ConfigParser()
    params.read(str(args.parameters))

    parname = args.parname

    curves = {}

    print("Plotting...")

    for section in params.sections():

        short_name = parameters.ParameterName().from_section(section).name

        if parname.lower() in short_name:
            print("".join(["  [", section, "]"]))
            points = []

            for key, value in params.items(section):

                res = int(peaks.Peak(key).numbers["i"])
                split = value.split()
                value = float(split[0])

                try:
                    error = float(split[2])
                except ValueError:
                    error = 0.0

                points.append((res, value, error))

            curves[section] = zip(*points)

    _, axis = plt.subplots(figsize=(12, 5))

    axis.yaxis.grid(True)

    for section, (res, vals, errors) in curves.items():
        axis.errorbar(res, vals, yerr=errors, label=section, fmt=".", barsabove=True)

    plt.legend()
    plt.show()

    return 0



"""The peaks module contains the code for handling peak assignments and
resonances."""
import collections
import functools
import re

RE_PEAK_NAME = re.compile(
    """
        (^\s*|\-)
        (                                # group name
            (?P<symbol>\D?)              # one letter amino acid (optional)
            0*(?P<number>[0-9]+)         # residue number
            (?P<suffix>[abd-gi-mopr-z]*) # suffix (optional)
        )?
        (?P<nucleus>                     # nucleus name (e.g., CA, HG, ...)
            (?P<atom>[hncq])             # nucleus type
            [a-z0-9]*                    # nucleus name - nucleus type
        )?
    """,
    re.IGNORECASE | re.VERBOSE,
)

SPIN_NAMES = ("i", "s", "x")


@functools.total_ordering
class Peak:
    """Peak class."""

    def __init__(self, assignment=None):

        self._assignment = None

        if assignment is None:
            assignment = ""

        self.assignment = assignment

    @property
    def assignment(self):
        return self._assignment

    @assignment.setter
    def assignment(self, value):
        resonance_list = get_resonances(value.upper())
        self._resonances = collections.OrderedDict(
            (spin, res) for spin, res in zip("isx", resonance_list)
        )
        self._assignment = get_assignment(self._resonances.values())

    @property
    def names(self):
        return collections.OrderedDict(
            (
                ("i", self._resonances.get("i", {}).get("name", "")),
                ("s", self._resonances.get("s", {}).get("name", "")),
                (
                    "is",
                    get_assignment(
                        self._resonances[_] for _ in "is" if _ in self._resonances
                    ),
                ),
            )
        )

    @property
    def symbols(self):
        return collections.OrderedDict(
            (
                ("i", self._resonances.get("i", {}).get("symbol", "")),
                ("s", self._resonances.get("s", {}).get("symbol", "")),
            )
        )

    @property
    def atoms(self):
        return {key: res.get("atom", "") for key, res in self._resonances.items()}

    @property
    def nuclei(self):
        return {key: res.get("nucleus", "") for key, res in self._resonances.items()}

    @property
    def numbers(self):
        return {key: res.get("number", "") for key, res in self._resonances.items()}

    def __repr__(self):
        return self.assignment

    def __eq__(self, other):
        return self.assignment == other.assignment

    def __hash__(self):
        return hash(self.assignment)

    def __lt__(self, other):
        if isinstance(other, str):
            other = Peak(other)

        self_tuple = tuple(
            (resonance["nucleus"], int(resonance["number"]))
            for resonance in self._resonances.values()
            if resonance["number"]
        )

        other_tuple = tuple(
            (resonance["nucleus"], int(resonance["number"]))
            for resonance in other._resonances.values()
            if resonance["number"]
        )

        return self_tuple < other_tuple

    def __len__(self):
        return len(
            [resonance for resonance in self._resonances.values() if resonance["name"]]
        )

    def intersection(self, other):
        """TODO: method docstring."""

        assignments = {self.assignment, other.assignment}

        names = {resonance["name"] for resonance in self._resonances.values()}
        names.update({resonance["name"] for resonance in other._resonances.values()})

        groups = {resonance["group"] for resonance in self._resonances.values()}
        groups.update({resonance["group"] for resonance in other._resonances.values()})

        if len(assignments) == 1:
            assignment_new = assignments.pop()
        elif len(names) == 1:
            assignment_new = names.pop()
        elif len(groups) == 1:
            assignment_new = groups.pop()
        else:
            assignment_new = ""

        return Peak(assignment_new)


def get_resonances(assignment):
    """Get resonances from an assignment."""

    resonances = []
    last_resonance = None

    for match in re.finditer(RE_PEAK_NAME, assignment.lower()):

        resonance = match.groupdict()

        for key, value in resonance.items():
            if value is None:
                if last_resonance is not None:
                    resonance[key] = last_resonance[key]
                else:
                    resonance[key] = ""

        resonance["group"] = "".join(
            resonance.get(_, "") for _ in ("symbol", "number", "suffix")
        )
        resonance["name"] = "".join(resonance.get(_, "") for _ in ("group", "nucleus"))

        resonances.append(resonance)

        last_resonance = resonance

    return resonances


def get_assignment(resonances):
    """Get assignment from resonances."""
    parts = []
    last_group = None

    for resonance in resonances:
        if resonance["name"]:
            part = []
            if resonance["group"] is not None and resonance["group"] != last_group:
                part.append(resonance["group"])
            part.append(resonance["nucleus"])
            parts.append("".join(part))
            last_group = resonance["group"]

    return "-".join(parts)



"""The util module contains a variety of utility functions."""
import configparser
import sys


def read_cfg_file(filename):
    """Read and parse the experiment configuration file with configparser."""

    config = configparser.ConfigParser(inline_comment_prefixes=("#", ";"))
    config.optionxform = str

    try:
        out = config.read(str(filename))

        if not out and filename is not None:
            exit(f"\nERROR: The file '{filename}' is empty or does not exist!\n")

    except configparser.MissingSectionHeaderError:
        exit(f"\nERROR: You are missing a section heading in {filename:s}\n")

    except configparser.ParsingError:
        exit(
            "\nERROR: Having trouble reading your parameter file, did you"
            " forget '=' signs?\n{:s}".format(sys.exc_info()[1])
        )

    return config


def normalize_path(working_dir, filename):
    """Normalize the path of a filename relative to a specific directory."""

    path = filename

    if not path.is_absolute():
        path = working_dir / path

    return path.resolve()


def header1(string):
    """Print a formatted heading."""
    print(("\n".join(["", "", string, "=" * len(string), ""])))


def header2(string):
    """Print a formatted subheading."""
    print(("\n".join(["", string, "-" * len(string), ""])))



"""13C(methyl) - H to C CPMG

Measures methyl carbon chemical exchange recorded on site-specifically
13CH3-labeled proteins in a highly deuterated background. Magnetization is
initally anti-phase and is read out as in-phase. Because of the P-element only
even ncyc should be recorded. The calculation uses a 12x12 basis set:

[Cx(a), Cy(a), Cz(a), 2HxCz(a), 2HyCz(a), 2HzCz(a),
 Cx(b), Cy(b), Cz(b), 2HxCz(b), 2HyCz(b), 2HzCz(b)]

Off resonance effects are taken into account. The calculation is designed
explicitly for analyzing the Lewis Kay pulse sequence:

HtoC_CH3_exchange_*00_lek_ILV

Journal of Biomolecular NMR (2007) 38, 79-88
"""


import numpy as np

from chemex.experiments.cpmg import cpmg_profile
from chemex.spindynamics import basis, default

EXP_DETAILS = {
    "carrier": {"type": float},
    "time_t2": {"type": float},
    "pw90": {"type": float},
    "taub": {"default": 1.99e-3, "type": float},
    "time_equil": {"default": 0.0, "type": float},
}


class Profile(cpmg_profile.CPMGProfile):
    """TODO: class docstring."""

    def __init__(self, name, measurements, exp_details, model):
        super().__init__(name, measurements, exp_details, model)

        self.exp_details = self.check_exp_details(exp_details, expected=EXP_DETAILS)

        self.liouv = basis.Liouvillian(
            system="ixyzsz",
            state_nb=self.model.state_nb,
            atoms=self.peak.atoms,
            h_larmor_frq=self.conditions["h_larmor_frq"],
            equilibrium=False,
        )

        self.liouv.carrier_i = self.exp_details["carrier"]
        self.liouv.w1_i = 2.0 * np.pi / (4.0 * self.exp_details["pw90"])

        self.detect = self.liouv.detect["iz_a"]

        self.t_cps = {
            ncyc: self.exp_details["time_t2"] / (4.0 * ncyc) - self.exp_details["pw90"]
            for ncyc in self.ncycs[self.ncycs > 0]
        }

        self.t_neg = -2.0 * self.exp_details["pw90"] / np.pi

        self.delays = [
            self.t_neg,
            self.exp_details["time_equil"],
            self.exp_details["taub"],
        ]
        self.delays.extend(self.t_cps.values())

        self.map_names, self.params = default.create_params(
            basis=self.liouv,
            model=self.model,
            nuclei=self.peak.names,
            conditions=self.conditions,
            nh_constraints=True,
        )

        for name, full_name in self.map_names.items():
            if name.startswith(("dw", "r2_i_a")):
                self.params[full_name].set(vary=True)

    def calculate_unscaled_profile(self, **parvals):
        """Calculate the intensity in presence of exchange after a CEST block.

        Parameters
        ----------
        pb : float
            Fractional population of state B.
        kex_ab : float
            Exchange rates between states A and B in /s.
        dw_i_ab : float
            Chemical shift difference between states A and B in rad/s.
        r1_i_a : float
            Longitudinal relaxation rate of states A in /s.
        r2_i_a : float
            Transverse relaxation rate of state A in /s.
        dr2_i_ab : float
            Transverse relaxation rate difference between states A and B in /s.
        cs_i_a : float
            Resonance position of state A in ppm.

        Returns
        -------
        out : float
            Intensity after the CEST block

        """

        self.liouv.update(**parvals)

        # Calculation of all the needed propagators
        pulses = self.liouv.pulses_90_180_i()
        delays = dict(zip(self.delays, self.liouv.delays(self.delays)))

        # Simulate the CPMG block as function of ncyc
        mag0 = self.liouv.compute_mag_eq(term="2izsz", **parvals)

        profile = []

        p_element = (
            delays[self.exp_details["taub"]]
            @ pulses["90px"]
            @ self.liouv.perfect180["sx"]
            @ pulses["90px"]
            @ delays[self.exp_details["taub"]]
        )

        for ncyc in self.ncycs:

            if ncyc == 0:

                mag = (
                    self.detect
                    @ delays[self.exp_details["time_equil"]]
                    @ pulses["90py"]
                    @ p_element
                    @ pulses["90px"]
                    @ mag0
                )

            else:

                p_cpx = np.linalg.matrix_power(
                    delays[self.t_cps[ncyc]]
                    @ pulses["180px"]
                    @ delays[self.t_cps[ncyc]],
                    int(ncyc),
                )

                p_cpy = np.linalg.matrix_power(
                    delays[self.t_cps[ncyc]]
                    @ pulses["180py"]
                    @ delays[self.t_cps[ncyc]],
                    int(ncyc),
                )

                mag = (
                    self.detect
                    @ delays[self.exp_details["time_equil"]]
                    @ pulses["90py"]
                    @ delays[self.t_neg]
                    @ p_cpx
                    @ p_element
                    @ p_cpy
                    @ delays[self.t_neg]
                    @ pulses["90px"]
                    @ mag0
                )

            profile.append(np.float64(self.liouv.collapse(mag)))

        return np.asarray(profile)

    def ncycs_to_nu_cpmgs(self, ncycs=None):
        """Calculate the pulsing frequency, v(CPMG), from ncyc values."""

        if ncycs is None:
            ncycs = self.ncycs

        return ncycs / self.exp_details["time_t2"]



"""
Pure anti-phase CEST
====================

Analyzes chemical exchange during the CEST block. This is calculated using the
12x12, single spin matrix:

[ Ix(a), Iy(a), Iz(a), IxSz(a), IySz(a), IzSz(a),
  Ix(b), Iy(b), Iz(b), IxSz(b), IySz(b), IzSz(b) ]

Reference
=========

Sekhar, Rosenzweig, Bouvignies and Kay. PNAS (2016) 113:E2794-E2801

"""


import numpy as np

from chemex.experiments.cest import cest_profile
from chemex.spindynamics import basis, default

EXP_DETAILS = {
    "carrier": {"type": float},
    "time_t1": {"type": float},
    "b1_frq": {"type": float},
    "b1_inh": {"default": np.inf, "type": float},
    "b1_inh_res": {"default": 11, "type": int},
    "filter_offsets": {"default": 0.0, "type": float},
    "filter_bandwidths": {"default": 0.0, "type": float},
}


class Profile(cest_profile.CESTProfile):
    """Profile for pure in-phase CEST."""

    def __init__(self, name, measurements, exp_details, model):
        super().__init__(name, measurements, exp_details, model)

        self.exp_details = self.check_exp_details(exp_details, expected=EXP_DETAILS)

        self.liouv = basis.Liouvillian(
            system="ixyzsz",
            state_nb=self.model.state_nb,
            atoms=self.peak.atoms,
            h_larmor_frq=self.conditions["h_larmor_frq"],
            equilibrium=False,
        )

        self.liouv.w1_i = 2 * np.pi * self.exp_details["b1_frq"]
        self.liouv.w1_i_inh = self.exp_details["b1_inh"]
        self.liouv.w1_i_inh_res = self.exp_details["b1_inh_res"]

        self.carriers_i = self.b1_offsets_to_ppm()
        self.detect = self.liouv.detect["2izsz_a"]
        self.dephasing = self.exp_details["b1_inh"] == np.inf

        self.map_names, self.params = default.create_params(
            basis=self.liouv,
            model=self.model,
            nuclei=self.peak.names,
            conditions=self.conditions,
            hn_ap_constraints=True,
        )

        for name, full_name in self.map_names.items():
            if name.startswith(("dw", "r1_i_a", "r2")):
                self.params[full_name].set(vary=True)

    def calculate_unscaled_profile(self, b1_offsets=None, **parvals):
        """Calculate the CEST profile in the presence of exchange.

        TODO: Parameters
        ----------

        Returns
        -------
        out : float
            Intensity after the CEST block

        """

        reference = self.reference
        carriers_i = self.carriers_i

        if b1_offsets is not None:
            carriers_i = (
                self.exp_details["carrier"]
                + 2.0 * np.pi * b1_offsets / self.liouv.ppms["i"]
            )
            reference = [False for _ in b1_offsets]

        mag0 = self.liouv.compute_mag_eq(term="2izsz", **parvals)

        # As the CEST block is after t1 evolution, the excited state magnetization is set to 0.
        mag0[6:] = 0.0

        self.liouv.update(**parvals)

        profile = []

        for ref, carrier_i in zip(reference, carriers_i):
            self.liouv.carrier_i = carrier_i
            mag = mag0.copy()
            if not ref:
                mag = (
                    self.liouv.pulse_i(self.exp_details["time_t1"], 0.0, self.dephasing)
                    @ mag
                )
            mag = self.detect @ mag
            profile.append(np.float64(self.liouv.collapse(mag)))

        return np.asarray(profile)

    def filter_points(self, params=None):
        """Evaluate some criteria to know whether or not the point should be
        considered in the calculation."""

        cs_a = params[self.map_names["cs_i_a"]].value

        filter_offsets = np.asarray(self.exp_details["filter_offsets"]).reshape(-1)
        filter_bandwidths = np.asarray(self.exp_details["filter_bandwidths"]).reshape(
            -1
        )

        for offset, bandwidth in zip(filter_offsets, filter_bandwidths):
            nu_offsets = (
                (cs_a - self.exp_details["carrier"])
                * self.liouv.ppms["i"]
                / (2.0 * np.pi)
                - self.b1_offsets
                + offset
            )

            self.mask = np.logical_and(self.mask, abs(nu_offsets) > bandwidth * 0.5)

    def b1_offsets_to_ppm(self, b1_offsets=None):
        """Convert B1 offset from Hz to ppm."""
        if b1_offsets is None:
            b1_offsets = self.b1_offsets

        return (
            2.0 * np.pi * b1_offsets / self.liouv.ppms["i"]
            + self.exp_details["carrier"]
        )



"""The cpmg_profile module contains the code for handling CPMG profiles."""
import copy
from functools import lru_cache

import numpy as np

from chemex import peaks
from chemex.experiments import base_profile
from chemex.experiments.cpmg import plotting


class CPMGProfile(base_profile.BaseProfile):
    """CPMGProfile class."""

    def __init__(self, name, measurements, exp_details, model):
        super().__init__(name, exp_details, model)

        self.plot_data = plotting.plot_data
        self.calculate_unscaled_profile_cached = lru_cache(8)(
            self.calculate_unscaled_profile
        )

        self.peak = peaks.Peak(self.profile_name)

        self.ncycs = np.asarray(measurements["ncycs"], dtype=int)
        self.val = measurements["intensities"]
        self.err = measurements["intensities_err"]

        self.reference = self.ncycs == 0

        self.map_names = {}
        self.basis = None

    def calculate_residuals(self, params):
        """Calculate the residuals between the experimental and back-calculated
        values."""

        values = self.calculate_profile(params)
        residuals = (self.val - values) / self.err

        return residuals[self.mask]

    def calculate_unscaled_profile(self, **parvals):
        """Calculate the unscaled CPMG profile."""
        pass

    def calculate_scale(self, cal):
        """Calculate the scaling factor."""

        norm = sum((cal / self.err) ** 2)

        if norm:
            scale = sum(cal * self.val / self.err ** 2) / norm
        else:
            scale = 0.0

        return scale

    def calculate_profile(self, params):
        """Calculate the CMPG profile."""
        kwargs = {
            short_name: params[long_name].value
            for short_name, long_name in self.map_names.items()
        }

        values = self.calculate_unscaled_profile_cached(**kwargs)
        scale = self.calculate_scale(values)

        return values * scale

    def print_profile(self, params=None):
        """Print the CPMG profile."""
        output = []

        if params is not None:
            values = self.calculate_profile(params)
        else:
            values = self.val

        iter_vals = list(zip(self.ncycs, self.val, self.err, values))

        output.append(f"[{self.profile_name}]")
        output.append(
            "# {:>5s}   {:>17s} {:>17s} {:>17s}".format(
                "ncyc", "intensity (exp)", "uncertainty", "intensity (calc)"
            )
        )

        for ncyc, val, err, cal in iter_vals:
            line = f"  {ncyc:5.0f} = {val:17.8e} {err:17.8e}"

            if params is not None:
                line += f"{cal:17.8e}"
            else:
                line += "{:17s}".format("xxx")

            output.append(line)

        output.append("")
        output.append("")

        return "\n".join(output).upper()

    def filter_points(self, params=None):
        """Evaluate some criteria to know whether the point should be
        considered in the calculation or not."""

        self.mask = np.ones_like(self.val, dtype=np.bool)

    def make_mc_profile(self, params):
        """Make a CPMG profile for boostrap analysis."""

        profile = copy.copy(self)
        profile.val = (
            self.calculate_profile(params) + np.random.randn(len(self.val)) * self.err
        )

        return profile

    def make_bs_profile(self):
        """Make a CPMG profile for boostrap analysis."""
        indexes = np.array(range(len(self.val)))
        pool1 = indexes[self.reference]
        pool2 = indexes[np.logical_not(self.reference)]

        bs_indexes = []
        if pool1.size:
            bs_indexes.extend(np.random.choice(pool1, len(pool1)))
        bs_indexes.extend(np.random.choice(pool2, len(pool2)))

        bs_indexes = sorted(bs_indexes)

        profile = copy.copy(self)
        profile.ncycs = profile.ncycs[bs_indexes]
        profile.val = profile.val[bs_indexes]
        profile.err = profile.err[bs_indexes]

        profile.calculate_unscaled_profile_cached = lru_cache(5)(
            profile.calculate_unscaled_profile
        )

        return profile



""" The ref module contains the reference matrices and code for calculating the
    Liouvillian.

    Operator basis:
       {Ix, Iy, Iz, Sx, Sy, Sz,
        2IxSz, 2IySz, 2IzSx, 2IzSy,
        2IxSx, 2IxSy, 2IySx, 2IySy,
        2IzSz}
"""
import itertools

import numpy as np
from scipy import linalg
from scipy import stats

from chemex.spindynamics import constants

COMPONENTS = {
    name: index
    for index, name in enumerate(
        (
            "ix",  # 0
            "iy",  # 1
            "iz",  # 2
            "sx",  # 3
            "sy",  # 4
            "sz",  # 5
            "2ixsz",  # 6
            "2iysz",  # 7
            "2izsx",  # 8
            "2izsy",  # 9
            "2ixsx",  # 10
            "2ixsy",  # 11
            "2iysx",  # 12
            "2iysy",  # 13
            "2izsz",  # 14
        )
    )
}

SPIN_SYSTEMS = {
    "iz": np.array([2]),
    "ixy": np.array([0, 1]),
    "ixyz": np.array([0, 1, 2]),
    "ixyzsz": np.array([0, 1, 2, 6, 7, 14]),
    "ixyzsxyz": np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]),
    "ixysxy": np.array([10, 11, 12, 13]),
}

STATES_M = {
    "a": np.diag([1.0, 0.0, 0.0, 0.0]),
    "b": np.diag([0.0, 1.0, 0.0, 0.0]),
    "c": np.diag([0.0, 0.0, 1.0, 0.0]),
    "d": np.diag([0.0, 0.0, 0.0, 1.0]),
}

STATES_M["all"] = sum(STATES_M.values())
STATES_V = {state: np.diag(matrix).reshape(-1, 1) for state, matrix in STATES_M.items()}

POP_PAIRS = {
    component: [
        ("_".join((component, state)), "".join(("p", state)))
        for state in {"a", "b", "c", "d"}
    ]
    for component in COMPONENTS
}

N_SINGLE = len(COMPONENTS)
N_FULL = N_SINGLE * 4 + 1

ZEROS_M_SINGLE = np.zeros((N_SINGLE, N_SINGLE))
ZEROS_M_FULL = np.zeros((N_FULL, N_FULL))

ZEROS_V_SINGLE = np.zeros((N_SINGLE, 1))
ZEROS_V_FULL = np.zeros((N_FULL, 1))


class Liouvillian:
    """TODO"""

    def __init__(self, system, state_nb, atoms, h_larmor_frq, equilibrium=True):

        if atoms is None:
            atoms = {"i": "h"}

        self.ppms = {
            spin: 2.0 * np.pi * h_larmor_frq * constants.XI_RATIO[atom]
            for spin, atom in atoms.items()
        }

        self._vectors, matrices = build_basis(system, state_nb, equilibrium)
        self._matrices = add_cs_and_carrier(matrices, self.ppms)

        self.keys = self._matrices.keys()

        self.detect = {name: vector.T for name, vector in self._vectors.items()}

        self.perfect180 = make_perfect180(self._vectors)

        self._carrier_i = None
        self._carrier_s = None
        self._j_eff_i = None
        self._j_eff_i_weights = None
        self._l_carrier_i = None
        self._l_carrier_s = None
        self._l_j_eff_i = None
        self._w1_i = None
        self._w1_s = None
        self._w1_i_weights = 1.0
        self._w1_s_weights = 1.0
        self._w1_i_inh = 0.0
        self._w1_s_inh = 0.0
        self._w1_i_inh_res = 11
        self._w1_s_inh_res = 11

        self.update(cs_i_a=0.0)
        self.carrier_i = 0.0
        self.carrier_s = 0.0
        self.w1_i = 1e32
        self.w1_s = 1e32
        self.j_eff_i = 0.0
        self.j_eff_i_weights = 1.0

        zeros = np.zeros_like(self._matrices["kab"])
        self._rot90zp_i = linalg.expm(+0.5 * np.pi * self._matrices.get("w_i", zeros))
        self._rot90zm_i = linalg.expm(-0.5 * np.pi * self._matrices.get("w_i", zeros))
        self._rot90zp_s = linalg.expm(+0.5 * np.pi * self._matrices.get("w_s", zeros))
        self._rot90zm_s = linalg.expm(-0.5 * np.pi * self._matrices.get("w_s", zeros))

    @property
    def carrier_i(self):
        return self._carrier_i

    @carrier_i.setter
    def carrier_i(self, value):
        self._carrier_i = np.asarray(value)
        self._l_carrier_i = self._matrices.get("carrier_i", 0.0) * self._carrier_i

    @property
    def carrier_s(self):
        return self._carrier_s

    @carrier_s.setter
    def carrier_s(self, value):
        self._carrier_s = np.asarray(value)
        self._l_carrier_s = self._matrices.get("carrier_s", 0.0) * self._carrier_s

    @property
    def w1_i(self):
        return self._w1_i

    @w1_i.setter
    def w1_i(self, value):

        self._w1_i = value

        if self._w1_i_inh not in (0.0, np.inf) and self._w1_i_inh_res > 1:
            dist = np.linspace(-2.0, 2.0, self._w1_i_inh_res)
            w1_i_inh = self._w1_i_inh
        else:
            dist = np.array(0.0)
            w1_i_inh = 0.0

        dist = dist.reshape(-1, 1, 1)

        w1_i_dist = (dist * w1_i_inh + 1.0) * self._w1_i

        self._w1_i_weights = stats.norm.pdf(dist)
        self._w1_i_weights /= self._w1_i_weights.sum()

        self._l_w1x_i = self._matrices.get("w1x_i", 0.0) * w1_i_dist
        self._l_w1y_i = self._matrices.get("w1y_i", 0.0) * w1_i_dist

    @property
    def w1_i_inh(self):
        return self._w1_i_inh

    @w1_i_inh.setter
    def w1_i_inh(self, value):
        self._w1_i_inh = value
        self.w1_i = self._w1_i

    @property
    def w1_i_inh_res(self):
        return self._w1_i_inh_res

    @w1_i_inh_res.setter
    def w1_i_inh_res(self, value):
        self._w1_i_inh_res = value
        self.w1_i = self._w1_i

    @property
    def w1_s(self):
        return self._w1_s

    @w1_s.setter
    def w1_s(self, value):

        self._w1_s = value

        if self._w1_s_inh not in (0.0, np.inf) and self._w1_s_inh_res > 1:
            dist = np.linspace(-2.0, 2.0, self._w1_s_inh_res)
            w1_s_inh = self._w1_s_inh
        else:
            dist = np.array(0.0)
            w1_s_inh = 0.0

        dist = dist.reshape(-1, 1, 1)

        w1_s_dist = (dist * w1_s_inh + 1.0) * self._w1_s

        self._w1_s_weights = stats.norm.pdf(dist)
        self._w1_s_weights /= self._w1_s_weights.sum()

        self._l_w1x_s = self._matrices.get("w1x_s", 0.0) * w1_s_dist
        self._l_w1y_s = self._matrices.get("w1y_s", 0.0) * w1_s_dist

    @property
    def w1_s_inh(self):
        return self._w1_s_inh

    @w1_s_inh.setter
    def w1_s_inh(self, value):
        self._w1_s_inh = value
        self.w1_s = self._w1_s

    @property
    def w1_s_inh_res(self):
        return self._w1_s_inh_res

    @w1_s_inh_res.setter
    def w1_s_inh_res(self, value):
        self._w1_s_inh_res = value
        self.w1_s = self._w1_s

    @property
    def j_eff_i(self):
        return self._j_eff_i.reshape(-1)

    @j_eff_i.setter
    def j_eff_i(self, value):
        self._j_eff_i = np.asarray(value).reshape(-1, 1, 1, 1, 1)
        self._l_j_eff_i = self._matrices.get("j_eff_i", 0.0) * self._j_eff_i

    @property
    def j_eff_i_weights(self):
        return self._j_eff_i_weights.reshape(-1)

    @j_eff_i_weights.setter
    def j_eff_i_weights(self, value):
        self._j_eff_i_weights = np.asarray(value).reshape(-1, 1, 1, 1, 1)

    def compute_mag_eq(self, term="iz", **parvals):
        return sum(
            self._vectors.get(name1, 0.0) * parvals.get(name2, 0.0)
            for name1, name2 in POP_PAIRS[term]
        )

    def update(self, **parvals):
        self._l_free = sum(
            self._matrices[name] * parvals[name]
            for name in set(parvals) & set(self._matrices)
        )

    def collapse(self, vector):
        weights = (
            self._w1_i_weights * self._w1_s_weights * self._j_eff_i_weights
        ).reshape(-1, 1, 1)
        vector_ = vector.reshape(-1, *vector.shape[-2:])
        return (vector_ * weights).sum()

    def delays(self, times):
        liouv = self._l_free + self._l_carrier_i + self._l_carrier_s + self._l_j_eff_i
        return calculate_propagators(liouv, times)

    def pulse_i(self, times, phase, dephasing=False):
        l_w1_i = self._l_w1x_i * np.cos(phase * np.pi * 0.5) + self._l_w1y_i * np.sin(
            phase * np.pi * 0.5
        )

        liouv = (
            self._l_free
            + self._l_carrier_i
            + self._l_carrier_s
            + self._l_j_eff_i
            + l_w1_i
        )

        return calculate_propagators(liouv, times, dephasing)

    def pulses_90_180_i(self):
        pulses = {}
        liouv = (
            self._l_free
            + self._l_carrier_i
            + self._l_carrier_s
            + self._l_j_eff_i
            + self._l_w1x_i
        )
        t90 = 0.5 * np.pi / self._w1_i
        pulses["90px"] = calculate_propagators(liouv, t90)
        rot90zp, rot90zm = self._rot90zp_i, self._rot90zm_i
        pulses["90py"] = rot90zp @ pulses["90px"] @ rot90zm
        pulses["90mx"] = rot90zp @ pulses["90py"] @ rot90zm
        pulses["90my"] = rot90zp @ pulses["90mx"] @ rot90zm
        pulses["180px"] = pulses["90px"] @ pulses["90px"]
        pulses["180py"] = pulses["90py"] @ pulses["90py"]
        pulses["180mx"] = pulses["90mx"] @ pulses["90mx"]
        pulses["180my"] = pulses["90my"] @ pulses["90my"]
        return pulses

    def pulse_s(self, times, phase, dephasing=False):
        l_w1_s = self._l_w1x_s * np.cos(phase * np.pi * 0.5) + self._l_w1y_s * np.sin(
            phase * np.pi * 0.5
        )
        liouv = (
            self._l_free
            + self._l_carrier_i
            + self._l_carrier_s
            + self._l_j_eff_i
            + l_w1_s
        )
        return calculate_propagators(liouv, times, dephasing)

    def pulses_90_180_s(self):
        pulses = {}
        liouv = self._l_free + self._l_j_eff_i + self._l_w1x_s
        t90 = 0.5 * np.pi / self._w1_s
        rot90zp, rot90zm = self._rot90zp_s, self._rot90zm_s
        pulses["90px"] = calculate_propagators(liouv, t90)
        pulses["90py"] = rot90zp @ pulses["90px"] @ rot90zm
        pulses["90mx"] = rot90zp @ pulses["90py"] @ rot90zm
        pulses["90my"] = rot90zp @ pulses["90mx"] @ rot90zm
        pulses["180px"] = pulses["90px"] @ pulses["90px"]
        pulses["180py"] = pulses["90py"] @ pulses["90py"]
        pulses["180mx"] = pulses["90mx"] @ pulses["90mx"]
        pulses["180my"] = pulses["90my"] @ pulses["90my"]
        return pulses

    def pulse_is(self, times, phase_i, phase_s, dephasing=False):
        l_w1_i = self._l_w1x_i * np.cos(phase_i * np.pi * 0.5) + self._l_w1y_i * np.sin(
            phase_i * np.pi * 0.5
        )
        l_w1_s = self._l_w1x_s * np.cos(phase_s * np.pi * 0.5) + self._l_w1y_s * np.sin(
            phase_s * np.pi * 0.5
        )
        liouv = (
            self._l_free
            + self._l_carrier_i
            + self._l_carrier_s
            + self._l_j_eff_i
            + l_w1_i
            + l_w1_s
        )
        return calculate_propagators(liouv, times, dephasing)


def build_4st_is_spin_system():
    indexes = {
        "l": {
            "j": ((7, 9, 1, 4, 0, 3, 6, 8), (0, 3, 6, 8, 7, 9, 1, 4)),
            "w_i": ((1, 7, 12, 13, 0, 6, 10, 11), (0, 6, 10, 11, 1, 7, 12, 13)),
            "w_s": ((4, 9, 11, 13, 3, 8, 10, 12), (3, 8, 10, 12, 4, 9, 11, 13)),
            "w1x_i": ((2, 14, 8, 9, 1, 7, 12, 13), (1, 7, 12, 13, 2, 14, 8, 9)),
            "w1y_i": ((0, 10, 11, 6, 2, 8, 9, 14), (2, 8, 9, 14, 0, 10, 11, 6)),
            "w1x_s": ((5, 14, 6, 7, 4, 9, 11, 13), (4, 9, 11, 13, 5, 14, 6, 7)),
            "w1y_s": ((3, 10, 12, 8, 5, 6, 7, 14), (5, 6, 7, 14, 3, 10, 12, 8)),
        },
        "r": {
            "r2_i": ((0, 1), (0, 1)),
            "r1_i": ((2), (2)),
            "r2_s": ((3, 4), (3, 4)),
            "r1_s": ((5), (5)),
            "r2a_i": ((6, 7), (6, 7)),
            "r2a_s": ((8, 9), (8, 9)),
            "r2_mq": ((10, 11, 12, 13), (10, 11, 12, 13)),
            "r1a": ((14), (14)),
            "etaxy_i": ((6, 7, 0, 1), (0, 1, 6, 7)),
            "etaxy_s": ((8, 9, 3, 4), (8, 9, 3, 4)),
            "etaz_i": ((14, 2), (2, 14)),
            "etaz_s": ((14, 5), (5, 14)),
            "sigma": ((5, 0), (0, 5)),
            "mu_mq": ((10, 11, 12, 13), (13, 12, 11, 10)),
        },
        "k": {
            "kab": ((0, 1), (0)),
            "kba": ((1, 0), (1)),
            "kac": ((0, 2), (0)),
            "kca": ((2, 0), (2)),
            "kad": ((0, 3), (0)),
            "kda": ((3, 0), (3)),
            "kbc": ((1, 2), (1)),
            "kcb": ((2, 1), (2)),
            "kbd": ((1, 3), (1)),
            "kdb": ((3, 1), (3)),
            "kcd": ((2, 3), (2)),
            "kdc": ((3, 2), (3)),
        },
        "t": {"theta_i": (2), "theta_s": (5), "theta_is": (14)},
    }

    values = {
        "l": np.array([+1.0, +1.0, +1.0, +1.0, -1.0, -1.0, -1.0, -1.0]),
        "r": np.array([-1.0]),
        "k": np.array([-1.0, +1.0]),
        "t": np.array([+1.0]),
    }

    scales = {"j": np.pi}

    matrices = {}

    for kind in ("l", "r"):
        for name, index_ij in indexes[kind].items():
            matrix = ZEROS_M_SINGLE.copy()
            matrix[index_ij] = values[kind] * scales.get(name, 1.0)
            for state_name, state_matrix in STATES_M.items():
                if state_name != "all" or name.startswith(("w1", "w_")):
                    fname = "_".join([name, state_name])
                    fname = fname.replace("_all", "")
                    matrices[fname] = ZEROS_M_FULL.copy()
                    matrices[fname][:-1, :-1] = np.kron(state_matrix, matrix)

    for name, index_ij in indexes["k"].items():
        matrix = np.zeros((4, 4))
        matrix[index_ij] = values["k"]
        matrices[name] = ZEROS_M_FULL.copy()
        matrices[name][:-1, :-1] = np.kron(matrix, np.eye(15))

    for name, i in indexes["t"].items():
        vector = ZEROS_V_SINGLE.copy()
        vector[i] = values["t"]
        for state_name, state_vector in STATES_V.items():
            fname = "_".join([name, state_name])
            fname = fname.replace("_all", "")
            matrices[fname] = ZEROS_M_FULL.copy()
            matrices[fname][:-1, [-1]] = np.kron(state_vector, vector)

    for name in list(matrices):
        if name.startswith("w1") and name.endswith(("_a", "_b", "_c", "_d")):
            del matrices[name]

    vectors = {}

    for name, index_ij in COMPONENTS.items():
        vector = ZEROS_V_SINGLE.copy()
        vector[index_ij] = 1.0
        for state_name, state_vector in STATES_V.items():
            fname = "_".join([name, state_name])
            fname = fname.replace("_all", "")
            vectors[fname] = ZEROS_V_FULL.copy()
            vectors[fname][:-1, :] = np.kron(state_vector, vector)

    vectors["identity"] = ZEROS_V_FULL.copy()
    vectors["identity"][-1, 0] = +1.0

    return vectors, matrices


def build_basis(spin_system, state_nb, equilibrium=True):
    vectors, matrices = build_4st_is_spin_system()

    active_indexes = (
        SPIN_SYSTEMS[spin_system].reshape(1, -1)
        + N_SINGLE * np.arange(state_nb).reshape(-1, 1)
    ).reshape(-1)

    if equilibrium:
        active_indexes = np.concatenate((active_indexes, [-1]))

    vec_reduced = {}

    for name, vector in vectors.items():
        vector_ = vector[active_indexes]
        if vector_.any():
            vec_reduced[name] = vector_

    mesh = np.ix_(active_indexes, active_indexes)

    mat_reduced = {}

    for name, matrix in matrices.items():
        matrix_ = matrix[mesh]
        m_invalid = not matrix_.any()
        k_invalid = name.startswith("k") and matrix_.sum() != 0.0
        if not (m_invalid or k_invalid):
            mat_reduced[name] = matrix[mesh]

    return vec_reduced, mat_reduced


def add_cs_and_carrier(matrices, ppms):

    matrices_ = matrices.copy()

    for spin, state in itertools.product(("i", "s"), ("a", "b", "c", "d")):
        w_name = "_".join(("w", spin, state))
        cs_name = "_".join(("cs", spin, state))
        if w_name in matrices:
            matrices_[cs_name] = matrices[w_name] * ppms.get(spin, 1.0)

    for spin in ("i", "s"):
        w_name = "_".join(("w", spin))
        carrier_name = "_".join(("carrier", spin))
        j_eff_name = "_".join(("j_eff", spin))
        if w_name in matrices:
            matrices_[carrier_name] = (
                -1.0 * matrices_.get(w_name, 0.0) * ppms.get(spin, 1.0)
            )
            matrices_[j_eff_name] = np.pi * matrices_.get(w_name, 0.0)

    return matrices_


def calculate_propagators(liouvillian, delays, dephasing=False):
    """TODO: function docstring."""

    delays_ = np.asarray(delays).reshape(-1)
    shape = liouvillian.shape

    propagators = []

    for a_liouvillian in liouvillian.reshape(-1, *shape[-2:]):
        s, vr = linalg.eig(a_liouvillian)
        vri = linalg.inv(vr)

        if dephasing:
            sl = np.where(abs(s.imag) < 1e-6)[0]
            vr, s, vri = vr[:, sl], s[sl], vri[sl, :]

        d = np.asarray([np.diag(np.exp(s * t)) for t in delays_])

        propagators.append((vr @ d @ vri).real)

    propagators = np.asarray(propagators).swapaxes(0, 1).reshape(-1, *shape)

    return propagators


def make_perfect180(vectors):

    vect_size = list(vectors.values())[0].size

    perfect180 = {ptype: np.identity(vect_size) for ptype in ("ix", "iy", "sx", "sy")}

    for key in set(vectors) & set(COMPONENTS):
        if "ix" in key:
            perfect180["iy"] -= 2 * np.diag(vectors[key].reshape(-1))
        if "iy" in key:
            perfect180["ix"] -= 2 * np.diag(vectors[key].reshape(-1))
        if "iz" in key:
            perfect180["ix"] -= 2 * np.diag(vectors[key].reshape(-1))
            perfect180["iy"] -= 2 * np.diag(vectors[key].reshape(-1))
        if "sx" in key:
            perfect180["sy"] -= 2 * np.diag(vectors[key].reshape(-1))
        if "sy" in key:
            perfect180["sx"] -= 2 * np.diag(vectors[key].reshape(-1))
        if "sz" in key:
            perfect180["sx"] -= 2 * np.diag(vectors[key].reshape(-1))
            perfect180["sy"] -= 2 * np.diag(vectors[key].reshape(-1))

    return perfect180



"""Plot the CEST profiles."""
import contextlib

import numpy as np
from matplotlib import gridspec as gsp
from matplotlib import pyplot as plt
from matplotlib import ticker
from matplotlib.backends import backend_pdf

from chemex.experiments import plotting


def sigma_estimator(x):
    """Estimates standard deviation using median to exclude outliers.

    Up to 50% can be bad.

    """
    return np.median([np.median(abs(xi - np.asarray(x))) for xi in x]) * 1.1926


def compute_profiles(data_grouped, params):
    """Compute the CEST profiles used for plotting."""
    profiles = {}

    for peak, profile in data_grouped.items():
        mask_ref = profile.reference
        mask = np.logical_not(mask_ref)

        val_ref = np.mean(profile.val[mask_ref])

        b1_ppm_exp = profile.b1_offsets_to_ppm()[mask]
        mag_cal = profile.calculate_profile(params)[mask] / val_ref
        mag_exp = profile.val[mask] / val_ref
        mag_err = profile.err[mask] / np.absolute(val_ref)
        filtered = profile.mask[mask]

        b1_offsets_min, b1_offsets_max = plotting.set_lim(
            profile.b1_offsets[mask], 0.02
        )
        b1_offsets = np.linspace(b1_offsets_min, b1_offsets_max, 500)

        b1_ppm_fit = profile.b1_offsets_to_ppm(b1_offsets)
        mag_fit = profile.calculate_profile(params, b1_offsets) / val_ref

        cs_a = params.get(profile.map_names.get("cs_i_a"))
        cs_b = params.get(profile.map_names.get("cs_i_b"))
        cs_c = params.get(profile.map_names.get("cs_i_c"))
        cs_d = params.get(profile.map_names.get("cs_i_d"))

        profiles[peak] = (
            b1_ppm_exp,
            mag_cal,
            mag_exp,
            mag_err,
            b1_ppm_fit,
            mag_fit,
            cs_a,
            cs_b,
            cs_c,
            cs_d,
            filtered,
        )

    return profiles


def write_profile_fit(name, b1_ppm_fit, mag_fit, file_):
    """Write the fitted CEST profile."""

    file_.write("[{}]\n".format(name.upper()))

    file_.write("# {:>17s}   {:>17s}\n".format("OFFSET", "INTENSITY"))

    for b1_ppm_cal, mag_cal in zip(b1_ppm_fit, mag_fit):
        file_.write(f"  {b1_ppm_cal:17.8e} = {mag_cal:17.8e}\n")

    file_.write("\n")


def write_profile_exp(name, b1_ppm, mag_exp, mag_err, file_):
    """Write the experimental CEST profile."""

    file_.write("[{}]\n".format(name.upper()))

    file_.write(
        "# {:>17s}   {:>17s} {:>17s}\n".format("OFFSET", "INTENSITY", "UNCERTAINTY")
    )

    for b1_ppm_, mag_exp_, mag_err_ in zip(b1_ppm, mag_exp, mag_err):
        file_.write(f"  {b1_ppm_:17.8e} = {mag_exp_:17.8e} {mag_err_:17.8e}\n")

    file_.write("\n")


def plot_data(data, params, output_dir):
    """Write experimental and fitted data to a file and plot the CEST profiles.

    - *.exp: contains the experimental data
    - *.fit: contains the fitted data
    - *.pdf: contains the plot of experimental and fitted data

    """
    datasets = dict()

    for data_point in data:
        experiment_name = data_point.experiment_name
        datasets.setdefault(experiment_name, []).append(data_point)

    for experiment_name, dataset in datasets.items():
        basename = output_dir / experiment_name
        name_pdf = basename.with_suffix(".pdf")
        name_fit = basename.with_suffix(".fit")
        name_exp = basename.with_suffix(".exp")

        print((f"  * {name_pdf} [.fit, .exp]"))

        data_grouped = plotting.group_data(dataset)

        profiles = compute_profiles(data_grouped, params)

        with contextlib.ExitStack() as stack:
            file_pdf = stack.enter_context(backend_pdf.PdfPages(name_pdf))
            file_fit = stack.enter_context(name_fit.open("w"))
            file_exp = stack.enter_context(name_exp.open("w"))

            for peak in sorted(profiles):
                (
                    b1_ppm,
                    mag_cal,
                    mag_exp,
                    mag_err,
                    b1_ppm_fit,
                    mag_fit,
                    cs_a,
                    cs_b,
                    cs_c,
                    cs_d,
                    filtered,
                ) = profiles[peak]

                write_profile_fit(peak.assignment, b1_ppm_fit, mag_fit, file_fit)
                write_profile_exp(peak.assignment, b1_ppm, mag_exp, mag_err, file_exp)

                # Matplotlib #
                grid_spec = gsp.GridSpec(2, 1, height_ratios=[1, 4])

                ax1 = plt.subplot(grid_spec[0])
                ax2 = plt.subplot(grid_spec[1])

                cs_colors = list(
                    zip(
                        (cs_a, cs_b, cs_c, cs_d),
                        ("Blue", "Red", "Orange", "Deep Orange"),
                    )
                )

                for ax_ in (ax1, ax2):
                    for a_cs, color in cs_colors:
                        if a_cs is not None:
                            ax_.axvline(
                                a_cs.value,
                                color=plotting.PALETTE[color]["100"],
                                linestyle="-",
                                linewidth=1.0,
                                zorder=-100,
                            )

                    ax_.axhline(
                        0, color=plotting.PALETTE["Black"]["Text"], linewidth=0.5
                    )

                ########################

                ax2.plot(
                    b1_ppm_fit,
                    mag_fit,
                    linestyle="-",
                    color=plotting.PALETTE["Grey"]["700"],
                    zorder=2,
                )

                ax2.errorbar(
                    b1_ppm[filtered],
                    mag_exp[filtered],
                    mag_err[filtered],
                    fmt="o",
                    markeredgecolor=plotting.PALETTE["Red"]["500"],
                    ecolor=plotting.PALETTE["Red"]["500"],
                    markerfacecolor="None",
                    zorder=3,
                )

                unfiltered = np.logical_not(filtered)

                ax2.errorbar(
                    b1_ppm[unfiltered],
                    mag_exp[unfiltered],
                    mag_err[unfiltered],
                    fmt="o",
                    markeredgecolor=plotting.PALETTE["Red"]["100"],
                    ecolor=plotting.PALETTE["Red"]["100"],
                    markerfacecolor="None",
                    zorder=3,
                )

                xmin, xmax = plotting.set_lim(b1_ppm_fit, 0.05)
                mags = list(mag_exp) + list(mag_fit)
                ymin, ymax = plotting.set_lim(mags, 0.10)

                ax2.set_xlim(xmin, xmax)
                ax2.set_ylim(ymin, ymax)

                ax2.invert_xaxis()

                ax2.xaxis.set_major_locator(ticker.MaxNLocator(9))
                # ax2.yaxis.set_major_locator(ticker.MaxNLocator(6))

                ax2.xaxis.grid(False)

                ax2.set_xlabel(r"$\mathregular{B_1 \ position \ (ppm)}$")
                ax2.set_ylabel(r"$\mathregular{I/I_0}$")

                ########################

                deltas = np.asarray(mag_exp) - np.asarray(mag_cal)
                sigma = sigma_estimator(deltas)

                ax1.fill(
                    (xmin, xmin, xmax, xmax),
                    1.0 * sigma * np.asarray([-1.0, 1.0, 1.0, -1.0]),
                    fc=plotting.PALETTE["Black"]["Dividers"],
                    ec="none",
                )

                ax1.fill(
                    (xmin, xmin, xmax, xmax),
                    2.0 * sigma * np.asarray([-1.0, 1.0, 1.0, -1.0]),
                    fc=plotting.PALETTE["Black"]["Dividers"],
                    ec="none",
                )

                ax1.errorbar(
                    b1_ppm[filtered],
                    deltas[filtered],
                    mag_err[filtered],
                    fmt="o",
                    markeredgecolor=plotting.PALETTE["Red"]["500"],
                    ecolor=plotting.PALETTE["Red"]["500"],
                    markerfacecolor="None",
                    zorder=100,
                )

                ax1.errorbar(
                    b1_ppm[unfiltered],
                    deltas[unfiltered],
                    mag_err[unfiltered],
                    fmt="o",
                    markeredgecolor=plotting.PALETTE["Red"]["100"],
                    ecolor=plotting.PALETTE["Red"]["100"],
                    markerfacecolor="None",
                    zorder=100,
                )

                rmin, rmax = plotting.set_lim(deltas, 0.1)
                rmin = min([-3 * sigma, rmin - max(mag_err)])
                rmax = max([+3 * sigma, rmax + max(mag_err)])

                ax1.set_xlim(xmin, xmax)
                ax1.set_ylim(rmin, rmax)

                ax1.invert_xaxis()

                ax1.xaxis.set_major_locator(ticker.MaxNLocator(9))
                ax1.yaxis.set_major_locator(ticker.MaxNLocator(4))

                ax1.xaxis.set_major_formatter(ticker.NullFormatter())

                ax1.xaxis.grid(False)
                ax1.yaxis.grid(False)

                ax1.ticklabel_format(style="sci", scilimits=(0, 0), axis="y")

                ax1.set_title("{:s}".format(peak.assignment.upper()))
                ax1.set_ylabel("Residual")

                ########################

                for ax in (ax1, ax2):
                    ax.yaxis.set_ticks_position("left")
                    ax.xaxis.set_ticks_position("bottom")

                ########################

                file_pdf.savefig()
                plt.close()

                ########################

    return



"""
15N - Pure In-phase Nitrogen CPMG with 0013 phase cycle
=======================================================

Analyzes 15N chemical exchange in the presence of high power 1H CW decoupling
during the CPMG block. This keeps the spin system purely in-phase throughout,
and is calculated using the (3n)x(3n), single spin matrix, where n is the number
of states:

[ Ix(a), Iy(a), Iz(a),
  Ix(b), Iy(b), Iz(b),
   ... ]

This version is modified such that CPMG pulses are applied with [0013] phase cycle
as shown in Daiwen's paper. The cw decoupling on 1H is assumed to be
strong enough (> 15 kHz) such that perfect 1H decoupling can be achieved.

References
----------

Jiang et al. Journal of Magnetic Resonance (2015) 257
Hansen, Vallurupalli and Kay. Journal of Physical Chemistry B (2008) 112:5898-5904

Experimental parameters
-----------------------
  * h_larmor_frq (1H Larmor frequency, in MHz)
  * temperature  (sample temperature, in Celsius)
  * p_total      (optional: protein concentration, in M)
  * l_total      (optional: ligand concentration, in M)
  * time_t2      (CPMG relaxation delay in seconds)
  * carrier      (position of the 15N carrier during the CPMG period, in ppm)
  * pw90         (15N 90 degree pulse width of CPMG pulses, in seconds)
  * time_equil   (equilibration delay at the end of the CPMG period, in seconds)
  * ncyc_max     (maximum number of cycles)


Extra parameters
----------------
  * path         (directory of the profiles)
  * error        (= 'file': uncertainties are taken from the profile files
                  = 'auto': uncertainties are calculated from duplicates)

"""

from functools import reduce

import numpy as np

from chemex.experiments.cpmg import cpmg_profile
from chemex.spindynamics import basis, default

EXP_DETAILS = {
    "carrier": {"type": float},
    "time_t2": {"type": float},
    "pw90": {"type": float},
    "time_equil": {"default": 0.0, "type": float},
    "ncyc_max": {"type": int},
}

CP_PHASES = [
    [0, 0, 1, 3, 0, 0, 3, 1, 0, 0, 3, 1, 0, 0, 1, 3],
    [1, 3, 2, 2, 3, 1, 2, 2, 3, 1, 2, 2, 1, 3, 2, 2],
]


class Profile(cpmg_profile.CPMGProfile):
    """TODO: class docstring."""

    def __init__(self, name, measurements, exp_details, model):
        super().__init__(name, measurements, exp_details, model)

        self.exp_details = self.check_exp_details(exp_details, expected=EXP_DETAILS)

        self.time_t2 = self.exp_details["time_t2"]
        self.time_eq = self.exp_details["time_equil"]
        self.pw90 = self.exp_details["pw90"]
        self.ncyc_max = self.exp_details["ncyc_max"]

        # Set the liouvillian
        self.liouv = basis.Liouvillian(
            system="ixyz",
            state_nb=self.model.state_nb,
            atoms=self.peak.atoms,
            h_larmor_frq=self.conditions["h_larmor_frq"],
            equilibrium=False,
        )

        self.liouv.carrier_i = self.exp_details["carrier"]
        self.liouv.w1_i = 2.0 * np.pi / (4.0 * self.pw90)

        # Set the row vector for detection
        self.detect = self.liouv.detect["iz_a"]

        # Set the delays in the experiments
        ncycs = self.ncycs[~self.reference]
        self.tau_cps = dict(zip(ncycs, self.time_t2 / (4.0 * ncycs) - 0.75 * self.pw90))
        self.deltas = dict(zip(ncycs, self.pw90 * (self.ncyc_max - ncycs)))
        self.deltas[0] = self.pw90 * (self.ncyc_max - 1)
        self.t_pos2 = +4.0 * self.pw90 / np.pi
        self.t_neg = -2.0 * self.pw90 / np.pi
        self.delays = (
            [self.t_neg, self.t_pos2, self.time_eq]
            + list(self.tau_cps.values())
            + list(self.deltas.values())
        )

        # Set the phase cycling of the cpmg pulses
        self.phases = {
            ncyc: np.take(CP_PHASES, np.flip(np.arange(2 * ncyc)), mode="wrap", axis=1)
            for ncyc in ncycs
        }

        # Get the parameters this profile depends on
        self.map_names, self.params = default.create_params(
            basis=self.liouv,
            model=self.model,
            nuclei=self.peak.names,
            conditions=self.conditions,
        )

        for name, full_name in self.map_names.items():
            if name.startswith(("dw", "r2_i_a")):
                self.params[full_name].set(vary=True)

    def calculate_unscaled_profile(self, **parvals):
        """TODO: Write docstring"""

        self.liouv.update(**parvals)

        # Calculation of the propagators corresponding to all the delays
        delays = dict(zip(self.delays, self.liouv.delays(self.delays)))
        d_neg = delays[self.t_neg]
        d_pos2 = delays[self.t_pos2]
        d_eq = delays[self.time_eq]
        delta = {ncyc: delays[delay] for ncyc, delay in self.deltas.items()}

        # Calculation of the propagators corresponding to all the pulses
        pulses = self.liouv.pulses_90_180_i()
        p90 = np.array([pulses[name] for name in ["90px", "90py", "90mx", "90my"]])
        p180 = np.array([pulses[name] for name in ["180px", "180py", "180mx", "180my"]])

        # Getting the starting magnetization
        mag0 = self.liouv.compute_mag_eq(term="iz", **parvals)

        # Calculating the cpmg trains
        cp = {0: p180[[0, 3]] @ d_pos2 @ p180[[0, 1]]}

        for ncyc in set(self.ncycs[~self.reference]):
            tau_cp = delays[self.tau_cps[ncyc]]
            phase_cp = self.phases[ncyc]

            echo = tau_cp @ p180 @ tau_cp

            cp[ncyc] = [reduce(np.matmul, echo[phases]) for phases in phase_cp]
            cp[ncyc] = d_neg @ np.array(cp[ncyc]) @ d_neg

        # Make profile
        profile = [
            self.liouv.collapse(
                self.detect @ d_eq @ delta[ncyc] @ p90[3] @ cp[ncyc] @ p90[1] @ mag0
            )
            for ncyc in self.ncycs
        ]

        return np.asarray(profile)

    def ncycs_to_nu_cpmgs(self, ncycs=None):
        """Calculate the pulsing frequency, v(CPMG), from ncyc values."""
        if ncycs is None:
            ncycs = self.ncycs

        return ncycs / self.exp_details["time_t2"]



"""The dataset module contains the code for handling the experimental data."""
import configparser
import copy
import importlib
import operator
import pathlib
import sys

import numpy as np

from chemex import util
from chemex.experiments import base_profile


class DataSet:
    """DataSet class for handling experimental data."""

    def __init__(self, other=None):
        self.data = []
        self.ndata = 0
        self.chisq_ref = 1e32

        if isinstance(other, DataSet):
            self.data = copy.deepcopy(other.data)
            self.ndata = other.ndata

        elif isinstance(other, base_profile.BaseProfile):
            self.data.append(other)
            self.ndata = len(other)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        return self.data[key]

    def __iter__(self):
        for some_data in self.data:
            yield some_data

    def __add__(self, other):
        data_sum = DataSet(self)
        data_sum.data.extend(other.data)
        data_sum.ndata = self.ndata + other.ndata
        return data_sum

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        self.data.extend(other.data)
        self.ndata += other.ndata
        return self

    def append(self, profile):
        """Append data to an exisiting dataset."""
        if not isinstance(profile, base_profile.BaseProfile):
            raise TypeError
        self.data.append(profile)
        self.ndata += len(profile.val)

    def calculate_residuals(self, params, verbose=True, threshold=1e-3):
        """Calculate the residuals."""

        residuals = np.concatenate(
            [profile.calculate_residuals(params) for profile in self.data]
        )

        if verbose:
            chisq = sum(residuals ** 2)

            change = (chisq - self.chisq_ref) / self.chisq_ref

            if change < -threshold:
                nvarys = len([param for param in params.values() if param.vary])
                redchi = chisq / (self.ndata - nvarys)

                print(f"  * {chisq:.3e} / {redchi:.3e}")

                self.chisq_ref = chisq

        return residuals

    def write_to(self, params, path):
        """Write experimental and fitted profiles to a file."""
        datasets = dict()

        for profile in self.data:
            experiment_name = profile.experiment_name
            datasets.setdefault(experiment_name, list()).append(profile)

        for experiment_name, data in datasets.items():

            filename = path / experiment_name
            filename = filename.with_suffix(".dat")

            print(f"  * {filename}")

            with filename.open("w") as f:
                for profile in sorted(data, key=operator.attrgetter("peak")):
                    f.write(profile.print_profile(params=params))

    def add_dataset_from_file(self, filename, model=None, included=None, excluded=None):
        """Add data from a file to the dataset."""

        if model is None:
            model = "2st.pb_kex"

        print("{:<45s} ".format(str(filename)), end="")

        # Get the directory of the input file
        working_dir = filename.parent

        # Parse the experiment configuration file
        config = util.read_cfg_file(filename)

        try:
            # Read the experiment information
            details = dict(config.items("experiment"))
            experiment_type = details["type"]
            experiment_class = experiment_type.split(".")[0]

            # Read the experimental parameters
            details.update(
                {
                    key.lower(): val
                    for key, val in config.items("experimental_parameters")
                }
            )

            # Read the profile information (name, filename)
            filenames = {key.lower(): val for key, val in config.items("data")}

        except configparser.NoSectionError as error:
            sys.exit(f"    Reading aborted: {error}")

        except KeyError as error:
            sys.exit(
                "\nIn the section 'experiment' of {}, '{}' must be provided!".format(
                    filename, error
                )
            )

        try:
            # Read (optional) additional parameters
            details.update(
                {key.lower(): val for key, val in config.items("extra_parameters")}
            )

        except configparser.NoSectionError:
            pass

        path = pathlib.Path(details.get("path", "."))
        path = util.normalize_path(working_dir, path)

        try:
            reading = importlib.import_module(
                ".".join(["chemex.experiments", experiment_class, "reading"])
            )

        except ImportError:
            sys.exit(
                "The experiment '{}', referred in '{}' is not implemented.".format(
                    experiment_type, filename
                )
            )

        data, ndata = reading.read_profiles(
            path, filenames, details, model, included, excluded
        )

        self.data.extend(data)
        self.ndata += ndata

        print("{:<25s} {:<25d}".format(experiment_type, len(data)))

        return data

    def make_bs_dataset(self):
        """Create a new dataset to run a bootstrap simulation."""

        data_bs = DataSet()

        for profile in self.data:
            data_bs.append(profile.make_bs_profile())

        return data_bs

    def make_mc_dataset(self, params):
        """Create a new dataset to run a Monte-Carlo simulation."""

        data_mc = DataSet()

        for profile in self.data:
            data_mc.append(profile.make_mc_profile(params=params))

        return data_mc


def read_data(args):
    """Read experimental setup and data."""
    util.header1("Reading Experimental Data")

    data = DataSet()

    model = res_incl = res_excl = None

    if "model" in args:
        model = args.model

    if "res_incl" in args:
        res_incl = args.res_incl

    if "res_excl" in args:
        res_excl = args.res_excl

    if args.experiments:
        print(("{:<45s} {:<25s} {:<25s}".format("File Name", "Experiment", "Profiles")))
        print(("{:<45s} {:<25s} {:<25s}".format("---------", "----------", "--------")))

        for filename in args.experiments:
            filename_ = pathlib.Path(filename)
            data.add_dataset_from_file(filename_, model, res_incl, res_excl)

    if not data.data:
        sys.exit("\nNo data to fit!\n")

    return data



"""ChemEx is an analysis program for chemical exchange detected by NMR.

It is designed to take almost any kind of NMR data to aid the analysis,
but the principle techniques are CPMG relaxation dispersion and Chemical
Exchange Saturation Transfer.

"""

__all__ = ["__version__"]

try:
    from chemex._version import version as __version__
except ImportError:
    # broken installation, we don't even try
    # unknown only works because we do poor mans version compare
    __version__ = "unknown"



from setuptools import find_packages
from setuptools import setup

with open("README.md", "rb") as f:
    long_description = f.read().decode("utf-8")

setup(
    name="chemex",
    use_scm_version={"write_to": "chemex/_version.py"},
    description="ChemEx is an analysis program for chemical exchange detected by NMR.",
    long_description=long_description,
    author="Guillaume Bouvignies",
    author_email="gbouvignies@gmail.com",
    url="https://github.com/gbouvignies/chemex",
    license="3-Clause BSD",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="nmr protein dynamics chemical exchange cpmg cest relaxation data fitting",
    packages=find_packages(),
    install_requires=[
        "numpy>=1.0",
        "scipy>=1.0",
        "matplotlib>=2.0",
        "lmfit>=0.9.11",
        "asteval>=0.9.11",
    ],
    python_requires=">=3.6",
    entry_points={"console_scripts": ["chemex = chemex.chemex:main"]},
)



import itertools
import re
import string

import lmfit as lf
from scipy import constants as cst

from chemex import parameters

OTHER_SPIN = {"i": "s", "s": "i"}

STATES = "abcd"

RE_VARNAME = re.compile(
    "^(?P<name>.+?)(_(?P<spin>[is]{1,2}))?(_(?P<state>[abcd]{1,2}))?$"
)


def create_params(
    basis,
    model,
    nuclei=None,
    conditions=None,
    hn_ap_constraints=False,
    nh_constraints=False,
):
    """TODO"""

    fnames_k, params_k = create_params_k(model=model, conditions=conditions)
    fnames_lr, params_lr = create_params_lr(basis, conditions, nuclei)

    params_lr = set_thermal_factors(fnames_lr, params_lr, nuclei)

    if hn_ap_constraints:
        fnames_lr, params_lr = set_hn_ap_constraints(
            fnames_lr, params_lr, nuclei=nuclei, conditions=conditions
        )

    if nh_constraints:
        fnames_lr, params_lr = set_nh_constraints(fnames_lr, params_lr)

    fnames = dict(**fnames_k, **fnames_lr)

    # hack to add parameters depending on parameters not added yet
    params = lf.Parameters()
    params.add_many(*(lf.Parameter(fname) for fname in fnames.values()))
    params.update(params_lr)
    params.update(params_k)

    return fnames, params


def create_params_lr(basis, conditions, nuclei):
    names = get_default_names(basis)
    fnames = set_fnames(names, nuclei, conditions)
    params = buid_params(fnames)
    for name, fname in fnames.items():
        if name.endswith(("_b", "_c", "_d")):
            name_a = "".join((name[:-1], "a"))
            fname_a = fnames[name_a]
            params[fname].set(expr=fname_a)
    for state in "bcd":
        short_name = f"cs_{state}"
        short_name_expr = f"cs_a + dw_a{state}"
        parameters.set_param_expr(params, short_name, short_name_expr)
    return fnames, params


def get_default_names(basis):
    """TODO"""

    names = set()
    for name in basis._matrices:

        name_parsed = RE_VARNAME.match(name).groupdict()

        if not name.startswith(("k", "p", "w1", "carrier", "j_eff", "w_")):
            names.add(name)

            if name_parsed["name"] == "cs" and name_parsed["state"] != "a":
                name_dw = "dw_{spin}_a{state}".format(**name_parsed)
                names.add(name_dw)

    return names


def set_fnames(names=None, nuclei=None, conditions=None):
    """TODO"""

    if names is None:
        names = {}

    fnames = {}

    for name in names:
        name_parsed = RE_VARNAME.match(name).groupdict()

        short_name = "_".join(
            name_parsed[key]
            for key in ("name", "state")
            if name_parsed[key] is not None
        )

        definition = {"temperature": conditions.get("temperature", 0.0)}

        spin = name_parsed["spin"]

        if spin is None:
            spin = "is"

        definition["nuclei"] = nuclei[spin]

        if not name.startswith(("dw", "cs", "j")):
            definition["h_larmor_frq"] = conditions.get("h_larmor_frq", 1.0)

        fnames[name] = parameters.ParameterName(short_name, **definition).to_full_name()

    return fnames


def buid_params(fnames):
    """TODO"""

    params = lf.Parameters()

    for name, full_name in fnames.items():
        param = lf.Parameter(full_name, value=0.0, vary=False)

        if name.startswith("p"):
            param.set(min=0.0, max=1.0)
        elif name.startswith("k"):
            param.set(min=0.0)
        elif name.startswith("r2"):
            param.set(value=10.0, min=0.0)
        elif name.startswith("r1"):
            param.set(value=1.0, min=0.0)
        elif name == "j":
            param.set(value=-93.0)

        params[full_name] = param

    return params


def set_thermal_factors(fnames, params, nuclei):
    """TODO"""

    for spin, state in itertools.product("is", STATES):

        theta = f"theta_{spin}_{state}"

        if theta in fnames:

            r1_expr = f"r1_{spin}_{state}"
            sigma = f"sigma_{state}"
            pop = f"p{state}"

            terms = []

            terms.append(
                "{g_ratio} * {r1_expr}".format(
                    g_ratio=nuclei.atoms[spin]["g_ratio"], r1_expr=fnames[r1_expr]
                )
            )

            if sigma in fnames:
                terms.append(
                    "{g_ratio} * {sigma}".format(
                        g_ratio=nuclei.atoms[OTHER_SPIN[spin]]["g_ratio"],
                        sigma=fnames[sigma],
                    )
                )

            expr = "{pop} * ({sum})".format(pop=fnames[pop], sum=" + ".join(terms))

            params[fnames[theta]].set(expr=expr)

    for state in STATES:

        theta = f"theta_is_{state}"

        if theta in fnames:

            pop = f"p{state}"

            terms = []

            for spin in "is":

                etaz = f"etaz_{spin}_{state}"

                if etaz in fnames:
                    terms.append(
                        "{gamma} * {etaz}".format(
                            gamma=nuclei.atoms[spin]["g_ratio"], etaz=fnames[etaz]
                        )
                    )

            expr = "{pop} * ({sum})".format(pop=fnames[pop], sum=" + ".join(terms))

            params[fnames[theta]].set(expr=expr)

    return params


def set_hn_ap_constraints(fnames, params, nuclei=None, conditions=None):

    name_s = nuclei.get("s")

    if name_s is None:
        return fnames, params

    attributes = {"nuclei": name_s}
    attributes.update(conditions)

    names = (f"r1_s_{state}" for state in STATES if f"r1_i_{state}" in fnames)
    fnames, params = _add_params(fnames, params, names, attributes)

    expr_r1_i = "{{r1a_{state}}} - {{r1_s_{state}}}"
    expr_r2a_i = "{{r2_i_{state}}} - {{r1_s_{state}}}"

    settings = {}

    for state in STATES:

        r1_s = f"r1_s_{state}"
        r1_i = f"r1_i_{state}"
        r2a_i = f"r2a_i_{state}"

        if r1_s not in fnames:
            continue

        expr_r1_s = None
        if state != "a":
            expr_r1_s = "{r1a_a}".format(**fnames)

        settings[r1_s] = {"value": 1.0, "min": 0.0, "vary": False, "expr": expr_r1_s}
        settings[r1_i] = {"expr": expr_r1_i.format(state=state).format(**fnames)}
        settings[r2a_i] = {"expr": expr_r2a_i.format(state=state).format(**fnames)}

    fnames, params = _apply_settings(fnames, params, settings)

    return fnames, params


def set_nh_constraints(fnames, params):

    expr_r2a_i = "{{r2_i_{state}}} + {{r1a_{state}}} - {{r1_i_{state}}}"
    expr_r2a_s = "{{r2_s_{state}}} - {{r1_i_{state}}}"

    settings = {}

    settings.update(
        {
            f"r2a_i_{state}": {"expr": expr_r2a_i.format(state=state).format(**fnames)}
            for state in STATES
            if f"r2a_i_{state}" in fnames
        }
    )

    settings.update(
        {
            f"r2a_s_{state}": {"expr": expr_r2a_s.format(state=state).format(**fnames)}
            for state in STATES
            if f"r2a_s_{state}" in fnames
        }
    )

    fnames, params = _apply_settings(fnames, params, settings)

    return fnames, params


def create_params_k(model=None, conditions=None):
    """Update the experimental and fitting parameters depending on the model.

    """

    fnames = {}
    params = lf.Parameters()

    attributes = {
        key: conditions.get(key) for key in ("temperature", "p_total", "l_total")
    }

    states = string.ascii_lowercase[: model.state_nb]
    names = ["k{}{}".format(*pairs) for pairs in itertools.permutations(states, 2)]
    names.extend([f"p{state}" for state in states])

    fnames, params = _add_params(fnames, params, names, attributes)

    kinetic_models = {
        "2st.pb_kex": model_2st_pb_kex,
        "3st.pb_kex": model_3st_pb_kex,
        "4st.pb_kex": model_4st_pb_kex,
        "2st.eyring": model_2st_eyring,
        "3st.eyring": model_3st_eyring,
        "2st.binding": model_2st_binding,
    }

    kinetic_model = kinetic_models.get(model.name)

    if kinetic_model is None:
        print("Warning: The 'model' option should either be:")
        for model in sorted(kinetic_models):
            print(f"    - '{model}'")
        print("Set to the default value: '2st.pb_kex'.")
        kinetic_model = kinetic_models["2st.pb_kex"]

    fnames, params = kinetic_model(conditions, fnames, params)

    return fnames, params


def model_2st_pb_kex(conditions, fnames, params):

    attributes = {
        key: conditions.get(key) for key in ("temperature", "p_total", "l_total")
    }

    names = ("kex_ab",)

    fnames, params = _add_params(fnames, params, names, attributes)

    settings = {
        "kex_ab": {"value": 200.0, "min": 0.0, "vary": True},
        "pa": {"expr": "1.0 - {pb}".format(**fnames)},
        "pb": {"value": 0.05, "min": 0.0, "max": 1.0, "vary": True},
        "kab": {"expr": "{kex_ab} * {pb}".format(**fnames)},
        "kba": {"expr": "{kex_ab} * {pa}".format(**fnames)},
    }

    fnames, params = _apply_settings(fnames, params, settings)

    return fnames, params


def model_2st_eyring(conditions, fnames, params):

    attributes = {key: conditions.get(key) for key in ("p_total", "l_total")}
    names = ("dh_b", "ds_b", "dh_ab", "ds_ab")
    fnames, params = _add_params(fnames, params, names, attributes)

    thermo = _define_thermo(conditions.get("temperature"))

    settings = {
        "dh_b": {"value": 6.5e03, "vary": True},
        "ds_b": {"value": 1.0e10},
        "dh_ab": {"value": 6.5e04, "vary": True},
        "ds_ab": {"value": 1.0e10},
        "pa": {"expr": "{kba} / ({kba} + {kab})".format(**fnames)},
        "pb": {"expr": "{kab} / ({kba} + {kab})".format(**fnames)},
        "kab": {"expr": _get_k_from_h_s("ab").format(**fnames, **thermo)},
        "kba": {"expr": _get_k_from_h_s("ba").format(**fnames, **thermo)},
    }

    fnames, params = _apply_settings(fnames, params, settings)

    return fnames, params


def model_2st_binding(conditions, fnames, params):

    attributes = {key: conditions.get(key) for key in ("temperature",)}
    names = ("kd", "kon", "koff")
    fnames, params = _add_params(fnames, params, names, attributes)

    p_total = conditions.get("p_total", 0.0)
    l_total = conditions.get("l_total", 0.0)
    delta = l_total - p_total
    extra = {"delta": delta, "l_total": l_total}

    expr_kab = "{kon} * 0.5 * ({delta} - {kd} + sqrt(({delta} - {kd}) ** 2 + 4.0 * {kd} * {l_total}))"

    settings = {
        "kon": {"value": 1.0e7, "vary": True},
        "koff": {"value": 10.0, "vary": True},
        "kd": {"expr": "{koff} / {kon}".format(**fnames)},
        "pa": {"expr": "{kba} / ({kba} + {kab})".format(**fnames)},
        "pb": {"expr": "{kab} / ({kba} + {kab})".format(**fnames)},
        "kab": {"expr": expr_kab.format(**fnames, **extra)},
        "kba": {"expr": "{koff}".format(**fnames)},
    }

    fnames, params = _apply_settings(fnames, params, settings)

    return fnames, params


def model_3st_pb_kex(conditions, fnames, params):

    attributes = {
        key: conditions.get(key) for key in ("temperature", "p_total", "l_total")
    }
    names = ("kex_ab", "kex_ac", "kex_bc")
    fnames, params = _add_params(fnames, params, names, attributes)

    settings = {
        "pa": {"expr": "1.0 - {pb} - {pc}".format(**fnames)},
        "pb": {"value": 0.02, "min": 0.0, "max": 1.0, "vary": True},
        "pc": {"value": 0.02, "min": 0.0, "max": 1.0, "vary": True},
        "kex_ab": {"min": 0.0, "value": 200.0, "vary": True},
        "kex_ac": {"min": 0.0, "value": 0.0},
        "kex_bc": {"min": 0.0, "value": 200.0, "vary": True},
        "kab": {"expr": _get_k_from_kex_p("ab").format(**fnames)},
        "kba": {"expr": _get_k_from_kex_p("ba").format(**fnames)},
        "kac": {"expr": _get_k_from_kex_p("ac").format(**fnames)},
        "kca": {"expr": _get_k_from_kex_p("ca").format(**fnames)},
        "kbc": {"expr": _get_k_from_kex_p("bc").format(**fnames)},
        "kcb": {"expr": _get_k_from_kex_p("cb").format(**fnames)},
    }

    fnames, params = _apply_settings(fnames, params, settings)

    return fnames, params


def model_3st_eyring(conditions, fnames, params):

    attributes = {key: conditions.get(key) for key in ("p_total", "l_total")}
    names = ["dh_b", "dh_c", "dh_ab", "dh_ac", "dh_bc"]
    names.extend(["ds_b", "ds_c", "ds_ab", "ds_ac", "ds_bc"])
    fnames, params = _add_params(fnames, params, names, attributes)

    thermo = _define_thermo(conditions.get("temperature"))

    expr_pa = "{kba} * {kca} / ({kba} * {kca} + {kab} * {kca} + {kac} * {kba})"
    expr_pb = "{kab} * {kcb} / ({kab} * {kcb} + {kba} * {kcb} + {kbc} * {kab})"
    expr_pc = "{kbc} * {kac} / ({kbc} * {kac} + {kcb} * {kab} + {kca} * {kbc})"

    settings = {
        "dh_b": {"value": 6.5e03, "vary": True},
        "dh_c": {"value": 6.5e03, "vary": True},
        "dh_ab": {"value": 6.5e04, "vary": True},
        "dh_bc": {"value": 6.5e04, "vary": True},
        "dh_ac": {"value": 1.0e10},
        "ds_b": {"value": 1.0e10},
        "ds_c": {"value": 1.0e10},
        "ds_ab": {"value": 1.0e10},
        "ds_bc": {"value": 1.0e10},
        "ds_ac": {"value": 1.0e10},
        "kab": {"expr": _get_k_from_h_s("ab").format(**fnames, **thermo)},
        "kba": {"expr": _get_k_from_h_s("bs").format(**fnames, **thermo)},
        "kac": {"expr": _get_k_from_h_s("ac").format(**fnames, **thermo)},
        "kca": {"expr": _get_k_from_h_s("ca").format(**fnames, **thermo)},
        "kbc": {"expr": _get_k_from_h_s("bc").format(**fnames, **thermo)},
        "kcb": {"expr": _get_k_from_h_s("cb").format(**fnames, **thermo)},
        "pa": {"expr": expr_pa.format(**fnames)},
        "pb": {"expr": expr_pb.format(**fnames)},
        "pc": {"expr": expr_pc.format(**fnames)},
    }

    fnames, params = _apply_settings(fnames, params, settings)

    return fnames, params


def model_4st_pb_kex(conditions, fnames, params):

    attributes = {
        key: conditions.get(key) for key in ("temperature", "p_total", "l_total")
    }
    names = ("kex_ab", "kex_ac", "kex_ad", "kex_bc", "kex_bd", "kex_cd")
    fnames, params = _add_params(fnames, params, names, attributes)

    settings = {
        "pa": {"expr": "1.0 - {pb} - {pc} - {pd}".format(**fnames)},
        "pb": {"value": 0.02, "min": 0.0, "max": 1.0, "vary": True},
        "pc": {"value": 0.02, "min": 0.0, "max": 1.0, "vary": True},
        "pd": {"value": 0.02, "min": 0.0, "max": 1.0, "vary": True},
        "kex_ab": {"min": 0.0, "value": 200.0, "vary": True},
        "kex_ac": {"min": 0.0, "value": 0.0},
        "kex_ad": {"min": 0.0, "value": 0.0},
        "kex_bc": {"min": 0.0, "value": 200.0, "vary": True},
        "kex_bd": {"min": 0.0, "value": 0.0},
        "kex_cd": {"min": 0.0, "value": 200.0, "vary": True},
        "kab": {"expr": _get_k_from_kex_p("ab").format(**fnames)},
        "kba": {"expr": _get_k_from_kex_p("ba").format(**fnames)},
        "kac": {"expr": _get_k_from_kex_p("ac").format(**fnames)},
        "kca": {"expr": _get_k_from_kex_p("ca").format(**fnames)},
        "kad": {"expr": _get_k_from_kex_p("ad").format(**fnames)},
        "kda": {"expr": _get_k_from_kex_p("da").format(**fnames)},
        "kbc": {"expr": _get_k_from_kex_p("bc").format(**fnames)},
        "kcb": {"expr": _get_k_from_kex_p("cb").format(**fnames)},
        "kbd": {"expr": _get_k_from_kex_p("bd").format(**fnames)},
        "kdb": {"expr": _get_k_from_kex_p("db").format(**fnames)},
        "kcd": {"expr": _get_k_from_kex_p("cd").format(**fnames)},
        "kdc": {"expr": _get_k_from_kex_p("dc").format(**fnames)},
    }

    fnames, params = _apply_settings(fnames, params, settings)

    return fnames, params


def _add_params(fnames, params, names, attributes):
    for name in names:
        fname = _get_fullname(name, attributes=attributes)
        fnames[name] = fname
        params.add(name=fname, vary=False)

    return fnames, params


def _get_fullname(shortname, attributes=None):
    if attributes is None:
        attributes = {}
    return parameters.ParameterName(shortname, **attributes).to_full_name()


def _apply_settings(fnames, params, settings):

    settings_filtered = {key: val for key, val in settings.items() if key in fnames}

    for name, setting in settings_filtered.items():
        fname = fnames[name]
        params[fname]._delay_asteval = False
        params[fname].set(**setting)

    for name in settings_filtered:
        fname = fnames[name]
        params[fname]._delay_asteval = True

    return fnames, params


def _get_k_from_kex_p(states):
    kwargs = {
        "kex": "{{kex_{}{}}}".format(*sorted(states)),
        "p1": "{{p{}}}".format(states[0]),
        "psum": "{{p{}}} + {{p{}}}".format(*sorted(states)),
    }
    return "{kex} * {p1} / {psum} if {psum} else 0.0".format(**kwargs)


def _get_k_from_h_s(states):
    expr_dh = "{{dh_{}{}}}".format(*sorted(states))
    expr_ds = "{{ds_{}{}}}".format(*sorted(states))
    if states[0] != "a":
        expr_dh += " - {{dh_{}}}".format(states[0])
        expr_ds += " - {{ds_{}}}".format(states[0])
    return "{{kbt_h}} * exp(-(({expr_dh}) - {{tk}} * ({expr_ds})) / {{rt}})".format(
        expr_dh=expr_dh, expr_ds=expr_ds
    )


def _define_thermo(celcius):
    tk = celcius + 273.15
    kbt_h = cst.k * tk / cst.h
    rt = cst.R * tk
    thermo = {"tk": tk, "kbt_h": kbt_h, "rt": rt}
    return thermo
